//
//  EAGLView.m
//  VirtuAR
//
//  Created by Romain on 04/02/10.
//  Copyright __MyCompanyName__ 2010. All rights reserved.
//

#import "EAGLView.h"

#import "VARWorld.h"

#define FPS_REPORT_PERIOD	5.0 // In seconds

@implementation EAGLView

@synthesize animating;
@synthesize renderer;
@dynamic animationFrameInterval;

// You must implement this method
+ (Class) layerClass
{
    return [CAEAGLLayer class];
}

//The GL view is stored in the nib file. When it's unarchived it's sent -initWithCoder:
- (id) initWithCoder:(NSCoder*)coder
{    
    if ((self = [super initWithCoder:coder]))
	{
		[self setupView];
    }
	
    return self;
}

- (id) init
{    
    if (self = [super init])
	{
		[self setupView];
    }
	
    return self;
}

-(void)setupView
{
	self.opaque = YES;
	self.backgroundColor = [UIColor colorWithWhite:0.5 alpha:1.0];
	// Get the layer
	CAEAGLLayer *eaglLayer = (CAEAGLLayer *)self.layer;
	
	eaglLayer.opaque = TRUE;
	eaglLayer.drawableProperties = [NSDictionary dictionaryWithObjectsAndKeys:
									[NSNumber numberWithBool:FALSE], kEAGLDrawablePropertyRetainedBacking, kEAGLColorFormatRGBA8, kEAGLDrawablePropertyColorFormat, nil];
	
//	renderer = [[VARWorld alloc] init];
	
	animating = FALSE;
	displayLinkSupported = FALSE;
	animationFrameInterval = 4;
	displayLink = nil;
	animationTimer = nil;
	firstFrame = YES;
	
	// A system version of 3.1 or greater is required to use CADisplayLink. The NSTimer
	// class is used as fallback when it isn't available.
	NSString *reqSysVer = @"3.1";
	NSString *currSysVer = [[UIDevice currentDevice] systemVersion];
	if ([currSysVer compare:reqSysVer options:NSNumericSearch] != NSOrderedAscending)
		displayLinkSupported = TRUE;
}

- (void) drawView:(CADisplayLink *)sender
{
	NSTimeInterval realTimeStamp = [NSDate timeIntervalSinceReferenceDate];
	if(firstFrame)
	{
		firstFrame = NO;
		_nbFrames = 0;
		_lastFPSReport = realTimeStamp;
		_lastFrameTimeStamp = 0.0;
	}
	else
	{
		NSTimeInterval intervalBetweenFrames = realTimeStamp - _lastFrameTimeStamp;
		float framesElapsed = intervalBetweenFrames / (sender.duration * sender.frameInterval);
		if(framesElapsed > 1.5)
		{
//			NSLog(@"Draw loop missed frames: %f. Last frame length: %f", framesElapsed, _lastFrameLength);
		}
	}
	
	_nbFrames++;
	if((realTimeStamp - _lastFPSReport) > FPS_REPORT_PERIOD)
	{
		NSLog(@"Draw loop: %f fps", _nbFrames/(realTimeStamp - _lastFPSReport));
		_nbFrames = 0;
		_lastFPSReport = realTimeStamp;
	}
	
	_lastFrameTimeStamp = realTimeStamp;
	
    [renderer render:realTimeStamp];
	NSTimeInterval endFrame = [NSDate timeIntervalSinceReferenceDate];
	_lastFrameLength = endFrame - realTimeStamp;
}

- (void) layoutSubviews
{
	UIView *container = self.superview;
	CGRect bnds;
	if(container == nil)
	{
		bnds = [UIScreen mainScreen].applicationFrame;
	}
	else
	{
		bnds = container.bounds;
	}
	bnds.size.width = bnds.size.width * 1.0;
	bnds.size.height = bnds.size.height * 1.0;
	self.bounds = bnds;
	CGRect frm = self.frame;
	CGPoint cntr = self.center;
	NSLog(@"%f %f %f %f", frm.origin.x, frm.origin.y, cntr.x, cntr.y);
	bnds = self.bounds;
	NSLog(@"%f %f %f %f", bnds.origin.x, bnds.origin.y, bnds.size.width, bnds.size.height);
	frm.origin.x = 0;
	frm.origin.y = 0;
	self.frame = frm;
	[renderer resizeFromLayer:(CAEAGLLayer*)self.layer];
}

- (NSInteger) animationFrameInterval
{
	return animationFrameInterval;
}

- (void) setAnimationFrameInterval:(NSInteger)frameInterval
{
	// Frame interval defines how many display frames must pass between each time the
	// display link fires. The display link will only fire 30 times a second when the
	// frame internal is two on a display that refreshes 60 times a second. The default
	// frame interval setting of one will fire 60 times a second when the display refreshes
	// at 60 times a second. A frame interval setting of less than one results in undefined
	// behavior.
	if (frameInterval >= 1)
	{
		animationFrameInterval = frameInterval;
		
		if (animating)
		{
			[self stopAnimation];
			[self startAnimation];
		}
	}
}

- (void) startAnimation
{
	if (!animating)
	{
		if (displayLinkSupported)
		{
			// CADisplayLink is API new to iPhone SDK 3.1. Compiling against earlier versions will result in a warning, but can be dismissed
			// if the system version runtime check for CADisplayLink exists in -initWithCoder:. The runtime check ensures this code will
			// not be called in system versions earlier than 3.1.

			displayLink = [NSClassFromString(@"CADisplayLink") displayLinkWithTarget:self selector:@selector(drawView:)];
			[displayLink setFrameInterval:animationFrameInterval];
			[displayLink addToRunLoop:[NSRunLoop currentRunLoop] forMode:NSDefaultRunLoopMode];
		}
		else
			animationTimer = [NSTimer scheduledTimerWithTimeInterval:(NSTimeInterval)((1.0 / 60.0) * animationFrameInterval) target:self selector:@selector(drawView:) userInfo:nil repeats:TRUE];
		
		animating = TRUE;
	}
}

- (void)stopAnimation
{
	if (animating)
	{
		if (displayLinkSupported)
		{
			[displayLink invalidate];
			displayLink = nil;
		}
		else
		{
			[animationTimer invalidate];
			animationTimer = nil;
		}
		
		animating = FALSE;
	}
}

/*
-(void)touchesBegan:(NSSet *)touches withEvent:(UIEvent *)event
{
	[self.nextResponder touchesBegan:touches withEvent:event];
}

-(void)touchesMoved:(NSSet *)touches withEvent:(UIEvent *)event
{
	[self.nextResponder touchesMoved:touches withEvent:event];
}

- (void) touchesEnded: (NSSet *) touches withEvent: (UIEvent *) event 
{	
	[self.nextResponder touchesEnded: touches withEvent:event]; 
}

- (void) touchesCancelled: (NSSet *) touches withEvent: (UIEvent *) event 
{	
	[self.nextResponder touchesCancelled: touches withEvent:event]; 
}
*/
 
- (void) dealloc
{
    [renderer release];
	
    [super dealloc];
}

@end
