//
//  NSMutableArray+FIFOStack.m
//  videoGL
//
//  Created by Romain Criton on 18/08/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import "FIFOStack.h"

#define DEFAULT_CAPACITY	5

@interface FIFOStack()
-(void)_lazyCreateCondition;
@end

@implementation FIFOStack

@synthesize maxCapacity = _maxCapacity;
@synthesize lowThreshold = _lowThreshold;
@synthesize lowBufferTrigger = _lowBufferTrigger;

-(id)initWithCapacity:(NSUInteger)numItems
{
	self = [super init];
	if(self == nil) return self;
	
	_array = [[NSMutableArray alloc] initWithCapacity:numItems];
	_maxCapacity = numItems;
	_lowThreshold = 0;
	
	return self;
}

-(id)init
{
	return [self initWithCapacity:DEFAULT_CAPACITY];
}

-(uint)count
{
	if(_array != nil)
		return _array.count;
	else
		return 0;
}

-(id)dequeue
{
	if([_array count] == 0) return nil;
	
	id obj = [_array objectAtIndex:0];
	[[obj retain] autorelease];
	[_array removeObjectAtIndex:0];
	return obj;
}

-(void)enqueue:(id) obj
{
	if(obj == nil) return;
	
	[_array addObject:obj];
}

-(id)dequeueSync
{
	id obj = nil;
	[self _lazyCreateCondition];
	[_lowBufferTrigger lock];
	obj = [self dequeue];
	[_lowBufferTrigger unlock];
	return obj;
}

-(void)enqueueSync:(id) obj
{
	[self _lazyCreateCondition];
	[_lowBufferTrigger lock];
	[self enqueue:obj];
	[_lowBufferTrigger unlock];
}

-(void)enqueueWaitIfFull:(id) obj
{
	[self _lazyCreateCondition];
	
	[_lowBufferTrigger lock];
	
	if(self.count >= _maxCapacity)
	{
		[_lowBufferTrigger wait];
	}

	while(self.count > _lowThreshold)
	{
		[_lowBufferTrigger wait];
	}
	
	[self enqueue:obj];
	
	[_lowBufferTrigger unlock];
}

-(void)enqueueThenWaitIfFull:(id) obj
{
	[self _lazyCreateCondition];
	
	[_lowBufferTrigger lock];
	
	[self enqueue:obj];
	
	if(self.count >= _maxCapacity)
	{
		[_lowBufferTrigger wait];
	}
	
	while(self.count > _lowThreshold)
	{
		[_lowBufferTrigger wait];
	}
	
	[_lowBufferTrigger unlock];	
}

-(id)dequeueSignalIfLow
{
	id obj = nil;
	[self _lazyCreateCondition];
	[_lowBufferTrigger lock];
	obj = [self dequeue];
	if(self.count <= _lowThreshold)
	{
		[_lowBufferTrigger broadcast];
	}
	[_lowBufferTrigger unlock];
	
	return obj; 
}

-(void)push:(id)obj
{
	if(obj == nil) return;
	[_array insertObject:obj atIndex:0];
}

-(void)pushSync:(id)obj
{
	[self _lazyCreateCondition];
	[_lowBufferTrigger lock];
	[self push:obj];
	[_lowBufferTrigger unlock];
}

-(NSCondition *)lowBufferTrigger
{
	[self _lazyCreateCondition];
	return _lowBufferTrigger;
}

-(void)_lazyCreateCondition
{
	if(_lowBufferTrigger == nil)
	{
		_lowBufferTrigger = [[NSCondition alloc] init];
	}
}

-(void)dealloc
{
	[_lowBufferTrigger release];
	[super dealloc];
}
@end
