//
//  RootViewController.m
//  VirtuAR
//
//  Created by Romain on 09/02/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import "GLViewController.h"
#import "VARWorld.h"
#import "EAGLView.h"
#import "VARSphere.h"
#import "VARBitmapTexture.h"

#define kAccelerometerFrequency		15.0 // Hz
#define kFilteringFactor			0.1
#define FILTERING_FACTOR_HEADING	0.5

#define FPS_REPORT_PERIOD			5.0

@implementation GLViewController
@synthesize videoPath=_videoPath;

#define LERP(A,B,C) ((A)*(1.0-C)+(B)*C)

-(void)blitFrame:(MWVideoFrame *)frame fromMediaPlayer:(MWMediaPlayer *)player
{
	if(frame != nil)
	{
		[sprite.texture setDataFromRaw:(void *)frame.data width:frame.width height:frame.height];
	}
}

-(void)displayNextFrame:(NSTimer *)timer {
	NSTimeInterval startTime = [NSDate timeIntervalSinceReferenceDate];
	MWVideoFrame *frame = [_mediaPlayer.videoFrameBuffer dequeueSignalIfLow];

	if(frame != nil)
	{
		[sprite.texture setDataFromRaw:(void *)frame.data width:frame.width height:frame.height];
	}
	
	float frameTime = 1.0/([NSDate timeIntervalSinceReferenceDate]-startTime);
	if (lastFrameTime<0) {
		lastFrameTime = frameTime;
	} else {
		lastFrameTime = LERP(frameTime, lastFrameTime, 0.8);
	}
}

/*
 // The designated initializer.  Override if you create the controller programmatically and want to perform customization that is not appropriate for viewDidLoad.
- (id)initWithNibName:(NSString *)nibNameOrNil bundle:(NSBundle *)nibBundleOrNil {
    if (self = [super initWithNibName:nibNameOrNil bundle:nibBundleOrNil]) {
        // Custom initialization
    }
    return self;
}
*/

/*
// Implement loadView to create a view hierarchy programmatically, without using a nib.
- (void)loadView {
}
*/


// Implement viewDidLoad to do additional setup after loading the view, typically from a nib.
- (void)viewDidLoad {
	VARWorld *world = [[VARWorld alloc] init];
	((EAGLView *)self.view).renderer = world;
	[world release];
	locManager = [[CLLocationManager alloc] init];
	locManager.delegate = self; // Assign your custom delegate object
/*
	locManager.distanceFilter = 5.0;
	[locManager startUpdatingLocation];
*/
	//Setup location and heading
	if (locManager.headingAvailable)
	{
		locManager.headingFilter = 1.0;
		[locManager startUpdatingHeading];
	}
	//Setup Accelerometer reporting
	[[UIAccelerometer sharedAccelerometer] setUpdateInterval:(1.0 / kAccelerometerFrequency)];
	[[UIAccelerometer sharedAccelerometer] setDelegate:self];
	
	Vector2D pos;
	pos.x = 0.0;
	pos.y = 0.0;
	[world.camera setCartesianPos2DFromPointer:&pos];

/*
	sprite = [[VAR3DSprite alloc] init];
	[sprite setTextureFileName:@"default_tex"];
	[sprite setWidth:20.0];
	[sprite setHeight:10.0];
	[sprite setVectorUpWithX:0.0 Y:1.0 Z:0.0];
	[sprite setVectorFrontWithX:0.0 Y:0.0 Z:1.0];
	[sprite updateOrientationMatrix];
	[sprite setCam:[world camera]];
	[sprite setCartesianPosWithX:0.0 Y:12.0 Z:7.0];
	[world addObject:sprite];
	[sprite release];
*/
	sprite = [[VARSphere alloc] initWithRadius:30.0 precision:64];
	sprite.texture = [[[VARBitmapTexture alloc] initWithPNGImageFromResources:@"default_tex"] autorelease];
	[sprite setCam:[world camera]];
	[sprite setCartesianPosWithX:0.0 Y:0.0 Z:0.0];
	[sprite setVectorUpWithX:0.0 Y:1.0 Z:0.0];
	[sprite setVectorFrontWithX:0.0 Y:0.0 Z:1.0];
	[sprite updateOrientationMatrix];
	[world addObject:sprite];
	[sprite release];	
	
	MWMediaFile *file = [[[MWMediaFile alloc] initWithPath:self.videoPath] autorelease];
	_mediaPlayer = [[MWMediaPlayer alloc] initWithFile:file];
	_mediaPlayer.delegate = self;
	[_mediaPlayer play];
	
	lastFrameTime = -1;
	
/*
	[NSTimer scheduledTimerWithTimeInterval:1.0/15
									 target:self
								   selector:@selector(displayNextFrame:)
								   userInfo:nil
									repeats:YES];
*/
	
	[self.view layoutSubviews];
	[(EAGLView *)self.view startAnimation];
}

-(void)viewWillAppear:(BOOL)animated
{
	[super viewWillAppear:animated];
}

- (void)viewDidAppear:(BOOL)animated
{
	[super viewDidAppear:animated];
}

-(void)viewWillDisappear:(BOOL)animated
{
	[super viewWillDisappear:animated];
}

-(void)viewDidDisappear:(BOOL)animated
{
	[super viewDidDisappear:animated];
}

/*
// Override to allow orientations other than the default portrait orientation.
- (BOOL)shouldAutorotateToInterfaceOrientation:(UIInterfaceOrientation)interfaceOrientation {
    // Return YES for supported orientations
    return (interfaceOrientation == UIInterfaceOrientationPortrait);
}
*/

- (void)didReceiveMemoryWarning {
	// Releases the view if it doesn't have a superview.
    [super didReceiveMemoryWarning];
	
	// Release any cached data, images, etc that aren't in use.
}

- (void)viewDidUnload {
	// Release any retained subviews of the main view.
	// e.g. self.myOutlet = nil;
}


- (void)dealloc {
    [super dealloc];
}

-(void)touchesBegan:(NSSet *)touches withEvent:(UIEvent *)event
{
//	DLog(@"GLViewController: touchesBegan");
//	[[UIAccelerometer sharedAccelerometer] setDelegate:nil];
	[locManager stopUpdatingHeading];
	_touchView = YES;
    UITouch *aTouch = [[event touchesForView:self.view] anyObject];
    CGPoint loc = [aTouch locationInView:self.view];
	loc.x /= self.view.frame.size.width;
	loc.y /= self.view.frame.size.height;
	[((VARWorld *)(((EAGLView *)self.view).renderer)).camera setReferencePoint:loc];
}

-(void)touchesMoved:(NSSet *)touches withEvent:(UIEvent *)event
{
//	DLog(@"GLViewController: touchesMoved");
	_tracking = YES;
    UITouch *aTouch = [[event touchesForView:self.view] anyObject];
    CGPoint loc = [aTouch locationInView:self.view];
	loc.x /= self.view.frame.size.width;
	loc.y /= self.view.frame.size.height;
	[((VARWorld *)(((EAGLView *)self.view).renderer)).camera moveReferencePointTo:loc];
}

- (void) touchesEnded: (NSSet *) touches withEvent: (UIEvent *) event 
{	
//	DLog(@"GLViewController: touchesEnded");
	if(!_tracking)
	{
		// Single tap
		[locManager startUpdatingHeading];
		_touchView = NO;
	}
	else
	{
		_tracking = NO;
	}
//	[[UIAccelerometer sharedAccelerometer] setDelegate:self];
//	[locManager startUpdatingHeading];
//	_touchView = NO;
}

- (void) touchesCancelled: (NSSet *) touches withEvent: (UIEvent *) event 
{	
//	DLog(@"GLViewController: touchesCancelled");
	_tracking = NO;
//	[[UIAccelerometer sharedAccelerometer] setDelegate:self];
//	[locManager startUpdatingHeading];
//	_touchView = NO;
}

#pragma mark UIAccelerometerDelegate methods
- (void)accelerometer:(UIAccelerometer*)accelerometer didAccelerate:(UIAcceleration*)acceleration
{
	NSTimeInterval realTimeStamp = [NSDate timeIntervalSinceReferenceDate];
	if(_firstFrame)
	{
		_firstFrame = NO;
		_nbFrames = 0;
		_lastFPSReport = realTimeStamp;
		_lastFrameTimeStamp = 0.0;
	}
	else
	{
		NSTimeInterval intervalBetweenFrames = realTimeStamp - _lastFrameTimeStamp;
		float framesElapsed = intervalBetweenFrames * kAccelerometerFrequency;
		if(framesElapsed > 1.5)
		{
//			NSLog(@"Input loop missed frames: %f. Last frame length: %f", framesElapsed, _lastFrameLength);
		}
	}
	
	_nbFrames++;
	if((realTimeStamp - _lastFPSReport) > FPS_REPORT_PERIOD)
	{
		NSLog(@"Input loop: accel %f fps, heading %f fps", _nbFrames/(realTimeStamp - _lastFPSReport), _headingFrames/(realTimeStamp - _lastFPSReport));
		NSLog(@"Video frame rate: %f", lastFrameTime);
		_nbFrames = 0;
		_headingFrames = 0;
		_lastFPSReport = realTimeStamp;
	}
	
	//Use a basic low-pass filter to only keep the gravity in the accelerometer values
	accel.x = acceleration.x * kFilteringFactor + accel.x * (1.0 - kFilteringFactor);
	accel.y = acceleration.y * kFilteringFactor + accel.y * (1.0 - kFilteringFactor);
	accel.z = acceleration.z * kFilteringFactor + accel.z * (1.0 - kFilteringFactor);
	
	Vector3D *acc;
	Vector3D accxy;
	
	if(_touchView)
	{
		accxy.x = accel.x;
		accxy.y = accel.y;
		accxy.z = 0.0;
		acc = &accxy;
	}
	else
	{
		acc = &accel;
	}
	
	Vector2D pos;
	pos.x = 0.0;
	pos.y = 0.0;
	
	[(VARWorld *)[(EAGLView *)self.view renderer]	setLocation:&pos
				  altitude:0.0
				 direction:direction
			  acceleration:acc];
	
	NSTimeInterval endFrame = [NSDate timeIntervalSinceReferenceDate];
	_lastFrameLength = endFrame - realTimeStamp;
	_lastFrameTimeStamp = realTimeStamp;
}

#pragma mark CLLocationManagerDelegate
- (void)locationManager:(CLLocationManager*)manager didUpdateHeading:(CLHeading*)newHeading
{
	_headingFrames++;
	// If the accuracy is valid, go ahead and process the event.
/*
	if ((newHeading.headingAccuracy >= 0) && (newHeading.trueHeading >= 0))
	{
		direction = newHeading.trueHeading;
	}
*/
	if ((newHeading.headingAccuracy >= 0) && (newHeading.magneticHeading >= 0))
	{
		direction = newHeading.magneticHeading;
	}
	//	NSLog(@"newHeading: %f +/- %f", newHeading.trueHeading, newHeading.headingAccuracy);
}


- (void)locationManager:(CLLocationManager *)manager didUpdateToLocation:(CLLocation *)newLocation fromLocation:(CLLocation *)oldLocation
{

}


@end
