//
//  MWLinkedList.m
//  VirtuAR
//
//  Created by Romain on 11/02/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import "MWLinkedList.h"

@interface MWLinkedList()
-(void)sortListAscending;
-(void)sortListDescending;
-(void)insertSortedLinkAscending:(MWLink *)link;
-(void)insertSortedLinkDescending:(MWLink *)link;
-(void)addLinkToEmptyList:(MWLink *)link;
@end

@implementation MWLinkedList
@synthesize head;
@synthesize tail;
@synthesize cursor;
@synthesize count;
@synthesize sortType;

-(void)addLink:(MWLink *)link
{
	[self addAtTailLink:link];
}

-(void)addLinkToEmptyList:(MWLink *)link
{
	self.head = link;
	self.tail = link;
	link.next = nil;
	link.prev = nil;
	self.count = 1;
}

-(void)addAtHeadLink:(MWLink *)link;
{
	if(self.count == 0)
		[self addLinkToEmptyList:link];
	else
		[self insertLink:link beforeLink:self.head];
}

-(void)addAtTailLink:(MWLink *)link
{
	if(self.count == 0)
		[self addLinkToEmptyList:link];
	else
		[self insertLink:link afterLink:self.tail];
}

-(void)insertLink:(MWLink *)link atIndex:(uint)index
{
	if(index >= self.count)
		return;
	
	if(link == nil)
		return;
	
	uint i;
	MWLink *cur = self.head;
	for(i = 0; i < index; i++)
	{
		if(cur != nil)
			cur = cur.next;
		else
			break;
	}
	
	if(i == index)
		[self insertLink:link beforeLink:cur];
}

-(void)insertLink:(MWLink *)newLink beforeLink:(MWLink *)before
{
	if((newLink == nil) || (before == nil))
		return;
	
	newLink.next = before;
	newLink.prev = before.prev;
	if(before.prev != nil)
	{
		[before.prev setNext:newLink];
	}
	else
	{
		self.head = newLink;
	}
	before.prev = newLink;
	
	self.count += 1;
}
	
-(void)insertLink:(MWLink *)newLink afterLink:(MWLink *)after;
{
	if((newLink == nil) || (after == nil))
		return;
	
	newLink.next = after.next;
	newLink.prev = after;
	if(after.next != nil)
	{
		[after.next setPrev:newLink];
	}
	else
	{
		self.tail = newLink;
	}
	after.next = newLink;
	
	self.count += 1;
}

	
-(MWLink *)getLinkAtIndex:(uint)index
{
	if(index >= self.count)
		return nil;
	
	MWLink *cur = self.head;
	uint i;
	for(i = 0; i < index; i++)
	{
		if(cur != nil)
			cur = cur.next;
		else 
			break;
	}
	
	if(i == index)
		return cur;
	else
		return nil;
}
	
-(MWLink *)getLinkForObject:(id)obj
{
	if(obj == nil)
		return nil;
	
	MWLink *cur = self.head;
	while(cur != nil)
	{
		if(cur.payload == obj)
			return cur;
		
		cur = cur.next;
	}
	
	return nil;
}

-(void)addObject:(id)obj
{
	[self addAtTailObject:obj];
}
-(void)addAtHeadObject:(id)obj
{
	MWLink *newLink = [[MWLink alloc] initWithPayload:obj];
	[self addAtHeadLink:newLink];
	[newLink release];
}

-(void)addAtTailObject:(id)obj;
{
	MWLink *newLink = [[MWLink alloc] initWithPayload:obj];
	[self addAtTailLink:newLink];
	[newLink release];
}

-(void)insertObject:(id)obj atIndex:(uint)index
{
	MWLink *newLink = [[MWLink alloc] initWithPayload:obj];
	[self insertLink:newLink atIndex:index];
	[newLink release];
}

-(void)insertObject:(id)obj beforeObject:(id)before
{
	MWLink *linkBefore = [self getLinkForObject:before];
	
	if(linkBefore == nil)
		return;
	
	[self insertObject:obj beforeLink:linkBefore];
}

-(void)insertObject:(id)obj afterObject:(id)after
{
	MWLink *linkAfter = [self getLinkForObject:after];
	
	if(linkAfter == nil)
		return;
	
	[self insertObject:obj afterLink:linkAfter];
}
-(void)insertObject:(id)obj beforeLink:(MWLink *)before;
{
	MWLink *newLink = [[MWLink alloc] initWithPayload:obj];
	[self insertLink:newLink beforeLink:before];
}
	 
-(void)insertObject:(id)obj afterLink:(MWLink *)after;
{
	MWLink *newLink = [[MWLink alloc] initWithPayload:obj];
	[self insertLink:newLink afterLink:after];
}
-(id)getObjectAtIndex:(uint)index
{
	return [[self getLinkAtIndex:index] payload];
}

-(id)removeHead
{
	id obj = [[self.head payload] autorelease];
	[self removeLink:self.head];
	return obj;
}

-(id)removeTail
{
	id obj = [[self.tail payload] autorelease];
	[self removeLink:self.tail];
	return obj;
}

-(void)push:(id)obj
{
	[self addObject:obj];
}
-(id)pop
{
	return [self removeTail];
}

-(id)removeAtIndex:(uint)index
{
	id obj = nil;
	if(index >= self.count) return nil;
	
	MWLink *cur = self.head;
	uint i;
	for(i = 0; i < index; i++)
	{
		if(cur != nil)
			cur = cur.next;
		else
		{
			break;
		}
	}
	
	if(i == index)
	{
		obj = [cur.payload autorelease];
		[self removeLink:cur];
	}
	return obj;
}

-(void)removeLink:(MWLink *)link
{
	if(link == nil) return;
	
	if(link.prev != nil)
		[link.prev setNext:link.next];
	else if(self.head == link)
		self.head = link.next;
	
	if(link.next != nil)
		[link.next setPrev:link.prev];
	else if(self.tail == link)
		self.tail = link.prev;
	
	self.count -= 1;
	if(self.cursor == link) [self rewindCursor];
}

-(void)removeObject:(id)obj
{
	MWLink *tmp = [self getLinkForObject:obj];
	if(tmp != nil)
		[self removeLink:tmp];
}

-(void)replaceObject:(id)oldObj withObject:(id)newObj
{
	MWLink *tmp = [self getLinkForObject:oldObj];
	if(tmp != nil)
		tmp.payload = newObj;
}

-(void)replaceLink:(MWLink *)oldLink   withLink:(MWLink *)newLink
{
	if((newLink == nil) || (oldLink == nil)) return;
	newLink.prev = oldLink.prev;
	newLink.next = oldLink.next;
	if(oldLink.prev != nil)
		[oldLink.prev setNext:newLink];
	else if(self.head == oldLink)
		self.head = newLink;
	
	if(oldLink.next != nil)
		[oldLink.next setPrev:newLink];
	else if(self.tail == oldLink)
		self.tail = newLink;
	
	if(self.cursor == oldLink)
		self.cursor = newLink;
}

-(void)replaceAtIndex:(uint)index withObject:(id)newObj
{
	MWLink *newLink = [[MWLink alloc] initWithPayload:newObj];
	[self replaceAtIndex:index withLink:newLink];
}
	
-(void)replaceAtIndex:(uint)index withLink:(MWLink *)newLink
{
	[self replaceLink:[self getLinkAtIndex:index] withLink:newLink];
}
	
-(void)rewindCursor
{
	self.cursor = self.head;
}

-(void)setCursorToTail
{
	self.cursor = self.tail;
}

-(MWLink *)nextLink
{
	MWLink *nextLink = self.cursor;
	if(nextLink != nil)
		self.cursor = nextLink.next;
	
	return nextLink;
}
	
-(MWLink *)prevLink
{
	MWLink *prevLink = self.cursor;
	if(prevLink != nil)
		self.cursor = prevLink.prev;
	
	return prevLink;
}

-(id)nextObject
{
	return [[self nextLink] payload];
}
	
-(id)prevObject
{
	return [[self prevLink] payload];
}

-(void)setSortType:(MWLLSortType)type
{
	switch(type)
	{
		case MWLL_SORTED_ASCENDING:
		case MWLL_SORTED_DESCENDING:
			sortType = type;
			[self sort];
			break;
		case MWLL_UNSORTED:
			sortType = type;
			break;
		default:
			sortType = MWLL_UNSORTED;
			break;
	}
}
-(void)sortListAscending
{
	MWLink *temp, *temp1, *temp2;
	id obj;
	
	temp = self.head;
	temp1 = self.head;
	
	for( ; temp.next != nil ; temp = temp.next)
	{
		for(;temp1.next != nil;)
		{
			temp2 = temp1;
			temp1 = temp1.next;
			
			if([temp2.payload respondsToSelector:@selector(compareToObject:)])
			{
				id<MWLinkedListCompare> cmp = temp2.payload;
				if([cmp compareToObject:temp1.payload] > 0)
				{
					obj = temp2.payload;
					temp2.payload = temp1.payload;
					temp1.payload = obj;
				}
			}
		}
		
		temp2 = self.head; 
		temp1 = self.head; // reinitialize after exiting first loop
	}
}
-(void)sortListDescending
{
	MWLink *temp, *temp1, *temp2;
	id obj;
	
	temp = self.head;
	temp1 = self.head;
	
	for( ; temp.next != nil ; temp = temp.next)
	{
		for(;temp1.next != nil;)
		{
			temp2 = temp1;
			temp1 = temp1.next;
			
			if([temp2.payload respondsToSelector:@selector(compareToObject:)])
			{
				id<MWLinkedListCompare> cmp = temp2.payload;
				if([cmp compareToObject:temp1.payload] < 0)
				{
					obj = temp2.payload;
					temp2.payload = temp1.payload;
					temp1.payload = obj;
				}
			}
		}
		
		temp2 = self.head; 
		temp1 = self.head; // reinitialize after exiting first loop
	}
}

-(void)sort
{
	switch(self.sortType)
	{
		case MWLL_SORTED_ASCENDING:
			[self sortListAscending];
			break;
		case MWLL_SORTED_DESCENDING:
			[self sortListDescending];
			break;
		case MWLL_UNSORTED:
			break;
		default:
			break;
	}
}	
-(void)insertSortedLink:(MWLink *)link
{
	switch(self.sortType)
	{
		case MWLL_SORTED_ASCENDING:
			[self insertSortedLinkAscending:link];
			break;
		case MWLL_SORTED_DESCENDING:
			[self insertSortedLinkDescending:link];
			break;
		case MWLL_UNSORTED:
			[self addLink:link];
			break;
		default:
			break;
	}
}

-(void)insertSortedObject:(id)obj
{
	MWLink *tmp = [[MWLink alloc] initWithPayload:obj];
	[self insertSortedLink:tmp];
	[tmp release];
}

-(void)insertSortedLinkAscending:(MWLink *)link
{
	if(self.sortType != MWLL_SORTED_ASCENDING)
		[self addLink:link];
	
	[self rewindCursor];
	id<MWLinkedListCompare> obj = link.payload;
	MWLink *tmp = [self nextLink];
	while(tmp != nil)
	{
		if([obj compareToObject:tmp.payload] < 0)
			break;
		
		tmp = [self nextLink];
	}
	
	if(tmp != nil)
		[self insertLink:link beforeLink:tmp];
	else
	{
		[self addAtTailLink:link];
	}
}

-(void)insertSortedLinkDescending:(MWLink *)link
{
	if(self.sortType != MWLL_SORTED_DESCENDING)
		[self addLink:link];
	
	[self rewindCursor];
	id<MWLinkedListCompare> obj = link.payload;
	MWLink *tmp = [self nextLink];
	while(tmp != nil)
	{
		if([obj compareToObject:tmp.payload] > 0)
			break;
		
		tmp = [self nextLink];
	}
	
	if(tmp != nil)
		[self insertLink:link beforeLink:tmp];
	else
	{
		[self addAtTailLink:link];
	}
}

-(void)pruneLinks
{
	self.head = nil;
	self.tail = nil;
	self.cursor = nil;
	self.count = 0;
}

-(void)dealloc
{
	[self pruneLinks];
	[super dealloc];
}

@end
