//
//  MWMediaFile.h
//  videoGL
//
//	Abstracts media file for reading and querying characteristics
//
//  Created by Romain Criton on 28/07/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import <Foundation/Foundation.h>
#include "libavformat/avformat.h"
#import "MWMediaPacket.h"

typedef struct
{
	UInt32	codec;
	UInt32	channels;
	double	sampleRate;
	UInt32	bitsPerSample;
	double	bitRate;
	BOOL	VBR;
}
MWMFAudioSettings;

typedef struct
{
	UInt32	codec;
	double	frameRate;
	UInt32	width;
	UInt32	height;
	double	bitRate;
}
MWMFVideoSettings;

typedef void MWMediaCodec;

@interface MWMediaFile : NSObject
{
	// Properties
	NSURL *_url;
	NSString *_filePath;
	UInt32 _fileFormat;
	NSTimeInterval _length;
	BOOL _audio;
	BOOL _video;
	MWMFAudioSettings _audioSettings;
	MWMediaCodec *_audioCodec;
	MWMediaCodec *_videoCodec;

	AVFormatContext *_pFormatCtx;
	int _videoStream;
	int _audioStream;
}

-(id)initWithPath:(NSString *)path;
-(id)initWithURL:(NSURL *)url;

-(MWMediaPacket *)nextPacket;

-(NSTimeInterval)getTimeStampForPacket:(MWMediaPacket *)packet;

@property (nonatomic, retain) NSURL *url;
@property (nonatomic, retain) NSString *filePath;
@property (nonatomic, readonly) UInt32 fileFormat;
@property (nonatomic, readonly) NSTimeInterval length;
@property (nonatomic, readonly) BOOL audio;
@property (nonatomic, readonly) BOOL video;
@property (nonatomic, readonly) MWMFAudioSettings audioSettings;
@property (nonatomic, readonly) MWMediaCodec *audioCodec;
@property (nonatomic, readonly) MWMediaCodec *videoCodec;

@end
