//
//  MWMediaFile.m
//  videoGL
//
//  Created by Romain Criton on 28/07/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import "MWMediaFile.h"

@interface MWMediaFile()
-(NSTimeInterval)pts2Timestamp:(int64_t)pts stream:(uint)stream;
@end



@implementation MWMediaFile
@synthesize url = _url;
@synthesize filePath = _filePath;
@synthesize fileFormat = _fileFormat;
@synthesize length = _length;
@synthesize audio = _audio;
@synthesize video = _video;
@synthesize audioCodec = _audioCodec;
@synthesize videoCodec = _videoCodec;
@synthesize audioSettings = _audioSettings;

-(id)initWithPath:(NSString *)path
{
	self = [super init];
	if(self == nil) return self;
	
	self.filePath = path;
    // Register all formats and codecs
    av_register_all();
	
    // Open video file
	int error;
    if((error = avformat_open_input(&_pFormatCtx, [path cStringUsingEncoding:NSASCIIStringEncoding], NULL, NULL))!=0)
        goto initError; // Couldn't open file
	
    // Retrieve stream information
    if(avformat_find_stream_info(_pFormatCtx, NULL)<0)
        goto initError; // Couldn't find stream information
	
    // Find the first video stream
    _videoStream=-1;
    for(int i=0; i<_pFormatCtx->nb_streams; i++)
	{ 
        if(_pFormatCtx->streams[i]->codec->codec_type==AVMEDIA_TYPE_VIDEO)
        {
            _videoStream=i;
			break;
        }
	}
	
    if(_videoStream==-1)
	{
		_video = NO;
		_videoCodec = NULL;
	}
	else
	{
		_video = YES;
		_videoCodec = _pFormatCtx->streams[_videoStream]->codec;
	}

	// Find the first audio stream
    _audioStream=-1;

	for(int i=0; i<_pFormatCtx->nb_streams; i++)
	{ 
		if(_pFormatCtx->streams[i]->codec->codec_type==AVMEDIA_TYPE_AUDIO)
		{
			_audioStream=i;
			break;
		}
	}
	
    if(_audioStream==-1)
	{
		_audio = NO;
		_audioCodec = NULL;
	}
	else
	{
		_audio = YES;
		_audioCodec = _pFormatCtx->streams[_audioStream]->codec;
	}
	
	return self;

initError:
	[self release];
	return nil;
}
-(id)initWithURL:(NSURL *)url
{
	return self;
}

-(void)dealloc
{
	[_url release];
	[_filePath release];
	[super dealloc];
}

-(MWMediaPacket *)nextPacket
{
	MWMediaPacket *pack = [[MWMediaPacket alloc] init];
	AVPacket *mediaPacket = malloc(sizeof(AVPacket));
	
	while(pack.data == NULL)
	{
		int status = av_read_frame(_pFormatCtx, mediaPacket);

		if(status >= 0)
		{
			if(mediaPacket->stream_index == _videoStream)
			{
				pack.video = YES;
				status = av_dup_packet(mediaPacket);
				pack.data = mediaPacket;
			}
			else if(mediaPacket->stream_index == _audioStream)
			{
				pack.video = NO;
				pack.data = mediaPacket;
			}
			else
			{
				av_free_packet(mediaPacket);
				continue;
			}
		}
		else
		{
			break;
		}
	}
	
	return [pack autorelease];
}

-(NSTimeInterval)pts2Timestamp:(int64_t)pts stream:(uint)stream
{
	AVRational timebase = _pFormatCtx->streams[stream]->time_base;
	NSTimeInterval timestamp = (NSTimeInterval)pts * (NSTimeInterval)timebase.num / ((NSTimeInterval)(timebase.den));
	return timestamp;
}

-(NSTimeInterval)getTimeStampForPacket:(MWMediaPacket *)packet
{
	if((packet == nil) || (packet.data == NULL)) return -1.0;
	
	return [self pts2Timestamp:((AVPacket *)packet.data)->pts stream:packet.video ? _videoStream : _audioStream];	
}

@end
