//
//  MWMediaPlayer.m
//  videoGL
//
//  Created by Romain Criton on 28/07/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import "MWMediaPlayer.h"

@interface MWMediaPlayer()
-(void)_threadMain:(id)arg;
-(void)_readLoop:(id)arg;
-(void)_videoDecodeLoop:(id)arg;
-(void)_primeLoop:(id)arg;
- (void)_blitFrame:(CADisplayLink *)sender;
- (void)_blitFrameNoAudio:(CADisplayLink *)sender;
@end


@implementation MWMediaPlayer

@synthesize file = _file;
@synthesize videoFrameBuffer = _videoFrameBuffer;
@synthesize delegate = _delegate;

-(id)init
{
	self = [super init];
	if(self == nil) return self;
	
	_readThread = [[NSThread alloc] initWithTarget:self
									selector:@selector(_readLoop:)
									  object:nil];
	_videoDecodeThread = [[NSThread alloc] initWithTarget:self
										  selector:@selector(_videoDecodeLoop:)
											object:nil];
	_lowBufferTrigger = [[NSCondition alloc] init];
	
	return self;
}

-(id)initWithFile:(MWMediaFile *)file
{
	self = [self init];
	if(self == nil) return self;
	
	self.file = file;
	
	return self;
}

-(void)dealloc
{
	[_readThread cancel];
	[_readThread release];
	[_videoDecodeThread cancel];
	[_videoDecodeThread release];
	[_audioDecoder release];
	[_videoDecoder release];
	[_videoFrameBuffer release];
	[_audioFrameBuffer release];
	
	self.file = nil;
	[super dealloc];
}

-(void)play
{
	if(_file.video)
	{
		_videoFrameBuffer = [[FIFOStack alloc] initWithCapacity:15];
		_videoFrameBuffer.lowThreshold = 5;
		_videoPacketBuffer = [[FIFOStack alloc] initWithCapacity:15];
		_videoPacketBuffer.lowThreshold = 5;
		_videoPacketBuffer.lowBufferTrigger = _lowBufferTrigger;
		_videoDecoder = [[MWVideoDecoder alloc] initWithCodec:_file.videoCodec];
		_useVideo = YES;
	}
	
	if(_file.audio)
	{
		_audioFrameBuffer = [[FIFOStack alloc] initWithCapacity:10];
		_audioDecoder = [[MWAudioDecoder alloc] initWithPath:_file.filePath inputBuffer:_audioFrameBuffer];
		_audioFrameBuffer.lowBufferTrigger = _lowBufferTrigger;
		_useAudio = YES;
	}

	[self _primeLoop:[NSNumber numberWithFloat:2.5]];
	if(_useAudio)
	{
		[_audioDecoder prime];
		_displayLink = [CADisplayLink displayLinkWithTarget:self selector:@selector(_blitFrame:)];
	}
	else
	{
		_displayLink = [CADisplayLink displayLinkWithTarget:self selector:@selector(_blitFrameNoAudio:)];
	}

	[_displayLink retain];
	// We set the smallest interval so that we can probe the duration of an interval
	[_displayLink setFrameInterval:1];
	_displayLinkLoop = [NSRunLoop currentRunLoop];
	[_readThread start];
	[_videoDecodeThread start];
	// Display Link tick is started in the Decode Loop
	// Audio is the started in the Display Link tick
}

- (void)_blitFrame:(CADisplayLink *)sender
{
	if(_blitCount == 0)
	{
		_blitCount++;
		return;
	}
	else if(_blitCount == 1)
	{
		_blitCount++;
		NSTimeInterval dur = sender.duration;
		DLog(@"MWMediaPlayer: Display Link tick duration is %f", dur);
		sender.frameInterval = 4;
		MWVideoFrame *frame = [_videoFrameBuffer dequeueSignalIfLow];
		if((frame != nil) && [_delegate respondsToSelector:@selector(blitFrame:fromMediaPlayer:)])
		{
			[_delegate blitFrame:frame fromMediaPlayer:self];
		}
		return;
	}
	else if(_blitCount == 2)
	{
		[_audioDecoder play];
	}
	
	MWVideoFrame *frame = nil;
	NSTimeInterval audioTime = [_audioDecoder getCurrentAudioTime];
	NSTimeInterval frameDuration = sender.duration * sender.frameInterval;
	
fetchFrame:
	frame = [_videoFrameBuffer dequeueSignalIfLow];
	
	if(frame != nil)
	{
		NSTimeInterval videoTime = frame.timestamp;
//		DLog(@"MWMediaPlayer: pts %f, audio %f", videoTime, audioTime);
		if((videoTime - audioTime) > frameDuration)
		{
			DLog(@"MWMediaPlayer: Video ahead of audio (%f > %f), holding off the frame", videoTime, audioTime);
			[_videoFrameBuffer pushSync:frame];
			return;
		}
		else if((audioTime - videoTime) > frameDuration)
		{
			DLog(@"MWMediaPlayer: Audio ahead of video (%f > %f), skipping a frame blit", audioTime, videoTime);
			goto fetchFrame;
		}
		
		if([_delegate respondsToSelector:@selector(blitFrame:fromMediaPlayer:)])
			[_delegate blitFrame:frame fromMediaPlayer:self];
	}
	else if(audioTime >= 0)
	{
		DLog(@"MWMediaPlayer: No frame to display");
		// TODO: pause the video for some time if blit is starved for too many consecutive frames
		// to allow the decoder to refill buffer
	}
}

- (void)_blitFrameNoAudio:(CADisplayLink *)sender
{
	if(_blitCount == 0)
	{
		_blitCount++;
		return;
	}
	else if(_blitCount == 1)
	{
		_blitCount++;
		NSTimeInterval dur = sender.duration;
		DLog(@"MWMediaPlayer: Display Link tick duration is %f", dur);
		sender.frameInterval = 4;
		MWVideoFrame *frame = [_videoFrameBuffer dequeueSignalIfLow];
		if((frame != nil) && [_delegate respondsToSelector:@selector(blitFrame:fromMediaPlayer:)])
		{
			[_delegate blitFrame:frame fromMediaPlayer:self];
		}
		return;
	}
	
	MWVideoFrame *frame = nil;
	
	frame = [_videoFrameBuffer dequeueSignalIfLow];
	
	if(frame != nil)
	{		
		if([_delegate respondsToSelector:@selector(blitFrame:fromMediaPlayer:)])
			[_delegate blitFrame:frame fromMediaPlayer:self];
	}
	else
	{
		DLog(@"MWMediaPlayer: No frame to display");
		// TODO: skip the missed frames if blit is starved
		// Of course if it is too starved, pause the play
	}
}

-(void)_threadMain:(id)arg
{
	[NSThread setThreadPriority:0.4];
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	
	//A runloop with no sources returns immediately from runMode:beforeDate:
	//That will wake up the loop and chew CPU. Add a dummy source to prevent
	//it.
	
	NSRunLoop *runLoop = [NSRunLoop currentRunLoop];
	
	NSMachPort *dummyPort = [[NSMachPort alloc] init];
	[runLoop addPort:dummyPort forMode:NSDefaultRunLoopMode];
	[dummyPort release];
	[pool release];
	
	while (!_readThread.isCancelled) 
	{
		NSAutoreleasePool *loopPool = [[NSAutoreleasePool alloc] init];
		[runLoop runMode:NSDefaultRunLoopMode beforeDate:[NSDate distantFuture]];
		[loopPool drain];
	}
}

-(void)_readLoop:(id)arg
{
	BOOL useAudio = _useAudio;
	BOOL useVideo = _useVideo;
//	[NSThread setThreadPriority:0.5];
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	
	//A runloop with no sources returns immediately from runMode:beforeDate:
	//That will wake up the loop and chew CPU. Add a dummy source to prevent
	//it.
	
	NSRunLoop *runLoop = [NSRunLoop currentRunLoop];
	
	NSMachPort *dummyPort = [[NSMachPort alloc] init];
	[runLoop addPort:dummyPort forMode:NSDefaultRunLoopMode];
	[dummyPort release];
	[pool release];
	
	MWMediaPacket *packet;
	
	DLog(@"MWMediaPlayer: readLoop start");
	
	while(!_readThread.isCancelled)
	{	
		NSAutoreleasePool *loopPool = [[NSAutoreleasePool alloc] init];
		packet = [_file nextPacket];

		if((packet == nil) || packet.data == NULL)
		{
			[_readThread cancel];
			continue;
		}
		
		if(packet.video && useVideo)
		{
			[_videoPacketBuffer enqueueSync:packet];
		}
		else if(useAudio)
		{
			[_audioFrameBuffer enqueueSync:packet];
		}
		
		[_lowBufferTrigger lock];
		if((_audioFrameBuffer.count >= _audioFrameBuffer.maxCapacity) && (_videoPacketBuffer.count >= _videoPacketBuffer.maxCapacity))
			[_lowBufferTrigger wait];
		
		while((_audioFrameBuffer.count > _audioFrameBuffer.lowThreshold) && (_videoPacketBuffer.count > _videoFrameBuffer.lowThreshold))
			[_lowBufferTrigger wait];
		
		[_lowBufferTrigger unlock];
		
		[runLoop runMode:NSDefaultRunLoopMode beforeDate:[NSDate distantPast]];
		[loopPool drain];
	}
	
	DLog(@"MWMediaPlayer: readLoop exit");
}

-(void)_videoDecodeLoop:(id)arg
{
//	[NSThread setThreadPriority:0.4];
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	
	//A runloop with no sources returns immediately from runMode:beforeDate:
	//That will wake up the loop and chew CPU. Add a dummy source to prevent
	//it.
	
	NSRunLoop *runLoop = [NSRunLoop currentRunLoop];
	
	NSMachPort *dummyPort = [[NSMachPort alloc] init];
	[runLoop addPort:dummyPort forMode:NSDefaultRunLoopMode];
	[dummyPort release];
	[pool release];
	
	MWMediaPacket *packet;
	
	DLog(@"MWMediaPlayer: videoDecodeLoop start");
	
	while(_videoFrameBuffer.count < _videoFrameBuffer.maxCapacity)
	{
		NSAutoreleasePool *loopPool = [[NSAutoreleasePool alloc] init];
		packet = [_videoPacketBuffer dequeueSignalIfLow];
		if(packet == nil)
		{
			DLog(@"MWMediaPlayer: video decode init phase starved");
			continue;
		}
		else if(!packet.video)
		{
			DLog(@"MWMediaPlayer: error, audio packet in video decode init phase");
			continue;
		}
		
		MWVideoFrame *pic = [_videoDecoder decodePacket:packet];
		if(pic != nil)
		{
			pic.timestamp = [_file getTimeStampForPacket:packet];
			[_videoFrameBuffer enqueueSync:pic];
		}
		
		[loopPool drain];
	}
	[_displayLink addToRunLoop:_displayLinkLoop forMode:NSDefaultRunLoopMode];
	// Audio is started in the Display Link tick
	
//	BOOL drop = NO;
	
	while(!_readThread.isCancelled)
	{	
		NSAutoreleasePool *loopPool = [[NSAutoreleasePool alloc] init];
		packet = [_videoPacketBuffer dequeueSignalIfLow];
		if(packet == nil)
		{
			DLog(@"MWMediaPlayer: video decode starved");
			continue;
		}
/*		
		AVPacket *avp = (AVPacket *)packet.data;

		if(drop)
		{
			if((avp->flags & PKT_FLAG_KEY) == 0)
			{
				DLog(@"MWMediaPlayer: video frames starved, drop frame");
				continue;
			}
			else
			{
				DLog(@"MWMediaPlayer: next keyframe, stop dropping frames");
			}
		}
*/
 
		MWVideoFrame *pic = [_videoDecoder decodePacket:packet];
		if(pic != nil)
		{
/*
			if(drop)
			{
				drop = NO;
			}
			else if(_videoFrameBuffer.count == 0)
			{
				drop = YES;
			}
*/
			pic.timestamp = [_file getTimeStampForPacket:packet];
			[_videoFrameBuffer enqueueWaitIfFull:pic];
		}

		MWVideoDecoderStats stats = _videoDecoder.stats;
		if(stats.instantFrameCount == 0)
		{
			if(stats.nbFrames != 0)
			{
				NSLog(@"MWMediaPlayer: decode time(min/avg/sigma/max): %f/%f/%f/%f", stats.minDuration, stats.avgDuration, sqrt(stats.sigmaSquaredDuration), stats.maxDuration);
				NSLog(@"MWMediaPlayer: FPS(theor/actual): %f/%f", stats.nbFrames / stats.totalDuration, stats.instantFrameRate);
			}
		}
		
		[runLoop runMode:NSDefaultRunLoopMode beforeDate:[NSDate distantPast]];
		[loopPool drain];
	}
	
	DLog(@"MWMediaPlayer: videoDecodeLoop exit");
}

-(void)_primeLoop:(id)arg
{
	if(arg == nil) return;

	BOOL useAudio = _useAudio;
	BOOL useVideo = _useVideo;
	MWMediaPacket *packet;
	NSTimeInterval maxTime = [arg floatValue];
	NSTimeInterval curTime = 0.0;
	
	packet = [_file nextPacket];
	while((curTime < maxTime) && (packet = [_file nextPacket]) && (packet.data != NULL))
	{
		curTime = [_file getTimeStampForPacket:packet];
		if(packet.video && useVideo)
		{
			[_videoPacketBuffer enqueue:packet];
		}
		else if(useAudio)
		{
			[_audioFrameBuffer enqueue:packet];
		}

	}
	DLog(@"MWMediaPlayer: primeLoop exit with video: %d, audio: %d", _videoPacketBuffer.count, _audioFrameBuffer.count);

	_audioFrameBuffer.maxCapacity = _audioFrameBuffer.count;
	_audioFrameBuffer.lowThreshold = 32;
}


@end
