//
//  VAR3DSprite.m
//  VirtuAR
//
//  Created by Romain on 05/02/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import "VAR3DSprite.h"
#import "VARGLutils.h"
#import "VARCamera.h"
#import "VARBitmapTexture.h"
#import "VARViewTexture.h"

// Size expressed relative to screen width
#define MIN_APPARENT_SIZE	0.35
#define MAX_APPARENT_SIZE	1.5

@implementation VAR3DSprite
@synthesize width;
@synthesize height;
@synthesize textureScaleX;
@synthesize textureScaleY;
@synthesize textureWrapS;
@synthesize textureWrapT;
@synthesize billboard;
@synthesize texture;

-(id)init
{
	if(self = [super init])
	{
		width = 1.0;
		height = 1.0;
		setVector3D(&vectorUp, 0.0, 0.0, 1.0);
		setVector3D(&vectorFront, 1.0, 0.0, 0.0);		
		setVector3D((Vector3D *)&(vertices[0]), -0.5, 0.5, 0.0);
		setVector3D((Vector3D *)&(vertices[1]), 0.5, 0.5, 0.0);
		setVector3D((Vector3D *)&(vertices[2]), -0.5, -0.5, 0.0);
		setVector3D((Vector3D *)&(vertices[3]), 0.5, -0.5, 0.0);
		setVector3D((Vector3D *)&(normals[0]), 0.0, 0.0, 1.0);
		setVector3D((Vector3D *)&(normals[1]), 0.0, 0.0, 1.0);
		setVector3D((Vector3D *)&(normals[2]), 0.0, 0.0, 1.0);
		setVector3D((Vector3D *)&(normals[3]), 0.0, 0.0, 1.0);
		setVector2D((Vector2D *)&(uvmap[0]), 0.0, 1.0);
		setVector2D((Vector2D *)&(uvmap[1]), 1.0, 1.0);
		setVector2D((Vector2D *)&(uvmap[2]), 0.0, 0.0);
		setVector2D((Vector2D *)&(uvmap[3]), 1.0, 0.0);
		textureWrapS = GL_CLAMP_TO_EDGE;
		textureWrapT = GL_CLAMP_TO_EDGE;
		billboard = YES;
		lastDraw = 0;
		scalingFactor = 1.0;
		bzero(scalingMatrix, sizeof(scalingMatrix));
		scalingMatrix[0][0] = scalingFactor;
		scalingMatrix[1][1] = scalingFactor;
		scalingMatrix[2][2] = scalingFactor;		
		scalingMatrix[3][3] = 1.0;
		
		[self updateOrientationMatrix];
	}
	
	return self;
}

-(void)setWidth:(GLfloat)w
{
	width = w;
	GLfloat half = w/2;
	vertices[0][0] = -half;
	vertices[1][0] = half;
	vertices[2][0] = -half;
	vertices[3][0] = half;
}

-(void)setHeight:(GLfloat)h
{
	height = h;
	GLfloat half = h/2;
	vertices[0][1] = half;
	vertices[1][1] = half;
	vertices[2][1] = -half;
	vertices[3][1] = -half;
}

-(void)setSize:(GLfloat)size
{
	self.width = size;
	self.height = size;
}

-(void)setSizeWidth:(GLfloat)w Height:(GLfloat)h
{
	self.width = w;
	self.height = h;
}

-(void)setTextureScaleX:(GLfloat)xscale
{
	textureScaleX = xscale;
	uvmap[1][0] = xscale;
	uvmap[3][0] = xscale;
}

-(void)setTextureScaleY:(GLfloat)yscale
{
	textureScaleY = yscale;
	uvmap[0][1] = yscale;
	uvmap[1][1] = yscale;
}

-(void)setTextureScale:(GLfloat)scale
{
	self.textureScaleX = scale;
	self.textureScaleY = scale;
}

-(void)setTextureFileName:(NSString *)filename
{
	texture = [[VARBitmapTexture alloc] initWithPNGImageFromResources:filename];
	texture.wrapS = textureWrapS;
	texture.wrapT = textureWrapT;
}

-(void)setTextureView:(UIView *)v
{
	texture = [[VARViewTexture alloc] initWithView:v];
	texture.wrapS = textureWrapS;
	texture.wrapT = textureWrapT;
}

-(void)setTextureWrapS:(GLint)wraps
{
	textureWrapS = wraps;
	[texture setWrapS:wraps];
}

-(void)setTextureWrapT:(GLint)wrapt
{
	textureWrapT = wrapt;
	[texture setWrapT:wrapt];
}

-(void)draw:(NSTimeInterval) timestamp
{
	if(texture == nil)
		return;
	
	glPushMatrix();
	
	// Translate object to its position
	glTranslatef(cartesianPos.x, cartesianPos.y, cartesianPos.z * sqrtf(scalingFactor));
	
	// Rotate object along Z axis so that it always faces the camera
	if(billboard)
		glMultMatrixf((GLfloat *)billboardMatrix);
	
	// Set correct object orientation:
	// - Up is Z
	// - Front is X
	glMultMatrixf((GLfloat*)orientationMatrix);
	
	// Scale object up or down to guarantee a minimum and maximum apparent size
	glScalef(scalingFactor, scalingFactor, scalingFactor);
//	glBlendFunc(GL_ONE, GL_SRC_COLOR);	
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	[texture bind];

	glVertexPointer(3, GL_FLOAT, 0, vertices);

    glNormalPointer(GL_FLOAT, 0, normals);
    glTexCoordPointer(2, GL_FLOAT, 0, uvmap);

    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
	
	glPopMatrix();
}

-(void)_updateDistanceToCam
{
	[super _updateDistanceToCam];
	VARCamera *camera = [self cam];
	Vector3D camPos = camera.cartesianPos;
	
	// Update Matrix used to rotate Sprite if Billboarding is enabled
	Vector3D delta;
	substractVectors3D(&camPos, &cartesianPos, &delta);
	normalizeVector3D(&delta);
	Vector3D col1 = { -delta.y, delta.x, 0.0 };
	Vector3D col2 = { 0.0, 0.0, 1.0 };
	
	matrixCreateFromColumns(billboardMatrix, &delta, &col1, &col2);
	
	// Update scaling factor to guarantee min and max sizes
	// That's the Viewport width expressed in world coordinates
	GLfloat viewSize = 2.0 * camera.zNear * tanf(DEGREES_TO_RADIANS(camera.fieldOfView) / 2.0);
	// Apparent width of the object
	GLfloat absoluteApparentSize = self.width * camera.zNear / distanceToCam;
	// Apparent width of the object expressed as a fraction of the Viewport width
	GLfloat apparentSize = absoluteApparentSize / viewSize;
	if(apparentSize > MAX_APPARENT_SIZE)
	{
		scalingFactor = MAX_APPARENT_SIZE / apparentSize;
	}
	else if(apparentSize < MIN_APPARENT_SIZE)
	{
		scalingFactor = MIN_APPARENT_SIZE / apparentSize;
	}
	else
	{
		scalingFactor = 1.0;
	}
	scalingMatrix[0][0] = scalingFactor;
	scalingMatrix[1][1] = scalingFactor;
	scalingMatrix[2][2] = scalingFactor;
}

-(void)dealloc
{
	[texture release];
	[super dealloc];
}
@end
