//
//  VARCamera.m
//  VirtuAR
//
//  Created by Romain on 05/02/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import "VARCamera.h"
#import "VARGLutils.h"

@interface VARCamera()
-(void)relativeViewPoint:(CGPoint *)viewPoint toFrustumPoint:(Vector3D *)frustumPoint;

@end


@implementation VARCamera

@synthesize viewSize;
@synthesize fieldOfView;
@synthesize zNear;
@synthesize zFar;
@synthesize accel;
@synthesize direction;
@synthesize positionChecksum;

-(id)init
{
	return [self initWithSizeX:320.0 Y:480.0];
}

-(id)init:(CGPoint)size
{
	return [self initWithSizeX:size.x Y:size.y];
}

-(id)initWithSizeX:(float)x Y:(float)y
{
	if(self = [super init])
	{
		[self setViewSizeWithX:x Y:y];
		zNear = 0.1;
		zFar = 2000.0;
		fieldOfView = 60.0;
		GLfloat size = zNear * tanf(DEGREES_TO_RADIANS(fieldOfView) / 2.0);
		frustumLeft = -size;
		frustumRight = size;
		frustumTop = size * viewSize.y / viewSize.x;
		frustumBottom = -frustumTop;
		
		_azimuth = 0.0;
		_elevation = 0.0;
		_dazimuth = 0.0;
		_delevation = 0.0;
		
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glFrustumf(frustumLeft, frustumRight, frustumBottom, frustumTop, zNear, zFar);
		glViewport(0, 0, viewSize.x, viewSize.y);
		
		//Make the OpenGL modelview matrix the default
		glMatrixMode(GL_MODELVIEW);
//		glTranslatef(0.0, -10.0, 3.0);
//		gluLookAt(10.0, -10.0, 10.0, 0.2, 0.0, 3.0, 0.0, 0.0, 1.0);
	}
	
	return self;
}

-(void)setZFar:(GLfloat)zfar
{
	zFar = zfar;

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustumf(frustumLeft, frustumRight, frustumBottom, frustumTop, zNear, zFar);
	glViewport(0, 0, viewSize.x, viewSize.y);
	
	//Make the OpenGL modelview matrix the default
	glMatrixMode(GL_MODELVIEW);

}

-(void)setViewSizeWithX:(float)x Y:(float)y
{
	viewSize.x = x;
	viewSize.y = y;
	glViewport(0, 0, viewSize.x, viewSize.y);
}

-(void)setAccelX:(GLfloat)x Y:(GLfloat)y Z:(GLfloat)z
{
	setVector3D(&accel, x, y, z);
}

-(void)setAccelFromPointer:(Vector3D *)acc
{
	copyVector3D(acc, &accel);
}

-(void)setReferencePoint:(CGPoint)point;
{
	_azimuth += _dazimuth;
	_elevation += _delevation;
	_dazimuth = 0.0;
	_delevation = 0.0;
	[self relativeViewPoint:&point toFrustumPoint:&_referencePoint];
	DLog(@"VARCamera: setReferencePoint relative(%f %f) frustum(%f %f %f) %f %f", point.x, point.y, _referencePoint.x, _referencePoint.y, _referencePoint.z, _azimuth, _elevation);
}

-(void)moveReferencePointTo:(CGPoint)point
{
	Vector3D currentPoint;
	[self relativeViewPoint:&point toFrustumPoint:&currentPoint];
	Vector3D up;
	negateVectors3D(&accel, &up);
	vectorAngles(&currentPoint, &_referencePoint, &up, &_dazimuth, &_delevation);
//	DLog(@"VARCamera: angles %f %f - delta %f %f", _azimuth, _elevation, _dazimuth, _delevation);
}

-(void)resetAzimuthAndElevation
{
	_azimuth = 0.0;
	_elevation = 0.0;
	_dazimuth = 0.0;
	_delevation = 0.0;
}

-(void)relativeViewPoint:(CGPoint *)viewPoint toFrustumPoint:(Vector3D *)frustumPoint
{
	frustumPoint->x = (-0.5 + viewPoint->x) * (frustumRight - frustumLeft);
	frustumPoint->y = (0.5 - viewPoint->y) * (frustumTop - frustumBottom);
	frustumPoint->z = -zNear;
}

-(void)draw:(NSTimeInterval) timestamp
{
	static GLfloat rot = 0;
	
    if (lastDraw)
    {
        NSTimeInterval timeSinceLastDraw = timestamp - lastDraw;
        rot +=  60 * timeSinceLastDraw;                
    }
    lastDraw = timestamp;
	
	GLfloat matrix[4][4], length;
	BOOL fakeAccel = NO;
	
	//Make sure we have a big enough acceleration vector
	length = vectorLength3D(&accel);
	
	if(length < 0.1)
	{
		setVector3D(&accel, 0.0, -1.0, 0.0);
		length = 1.0;
		fakeAccel = YES;
	}
	//Clear matrix to be used to rotate from the current referential to one based on the gravity vector

	bzero(matrix, sizeof(matrix));
	matrix[3][3] = 1.0;
	
	matrix[2][0] = -accel.x / length;
	matrix[2][1] = -accel.y / length;
	matrix[2][2] = -accel.z / length;
	
	matrix[0][0] = -accel.y / length;
	matrix[0][1] = accel.x / length;
	matrix[0][2] = 0.0;		
	normalizeVector3D((Vector3D *)&(matrix[0]));
	
	crossProduct((Vector3D *)&(matrix[2]), (Vector3D *)&(matrix[0]), (Vector3D *)&(matrix[1]));
	
	//Finally load matrix
	glMultMatrixf((GLfloat*)matrix);
	// TODO: when switching between portrait and landscape, direction changes by 90° on screen
//	glRotatef(_elevation + _delevation, matrix[0][0], matrix[0][1], matrix[0][2]);
	glRotatef(_elevation + _delevation, 1.0, 0.0, 0.0);
//	glRotatef(direction + _azimuth + _dazimuth, matrix[2][0], matrix[2][1], matrix[2][2]);
	glRotatef(direction + _azimuth + _dazimuth, 0.0, 0.0, 1.0);
	
	glTranslatef(-cartesianPos.x, -cartesianPos.y, -cartesianPos.z);
	
	if(fakeAccel)
	{
//		glRotatef(rot, 0.0, 0.0, 1.0);
	}
}

-(void)setCartesianPosWithX:(GLfloat)x Y:(GLfloat)y Z:(GLfloat)z
{
	positionChecksum = x + y + z;
	[super setCartesianPosWithX:x Y:y Z:z];
}

-(void)setCartesianPos2DWithX:(GLfloat)x Y:(GLfloat)y
{
	positionChecksum = x + y + self.cartesianPos.z;
	[super setCartesianPos2DWithX:x Y:y];
}

-(void)setCartesianPos2D:(Vector2D)pos
{
	positionChecksum = pos.x + pos.y + self.cartesianPos.z;
	[super setCartesianPos2D:pos];
}

-(void)setCartesianPosFromPointer:(Vector3D *)pos;
{
	positionChecksum = pos->x + pos->y + pos->z;
	[super setCartesianPosFromPointer:pos];
}

-(void)setCartesianPos2DFromPointer:(Vector2D *)pos
{
	positionChecksum = pos->x + pos->y + self.cartesianPos.z;
	[super setCartesianPos2DFromPointer:pos];
}

@end
