/*
 *  VARGLutils.h
 *  VirtuAR
 *
 *  Created by Romain on 04/02/10.
 *  Copyright 2010 __MyCompanyName__. All rights reserved.
 *
 */

#import <OpenGLES/ES1/gl.h>
#import <OpenGLES/ES1/glext.h>

#define ARE_FLOAT_EQUALS_WITH_PRECISION(x, y, p)	((x > y) ? ((x - y) < p) : ((y - x) < p))
#define DEFAULT_FLOAT_COMPARISON_PRECISION			0.0001
#define ARE_FLOAT_EQUALS(x, y)						ARE_FLOAT_EQUALS_WITH_PRECISION(x, y, DEFAULT_FLOAT_COMPARISON_PRECISION)

// Trigonometry
#define DEGREES_TO_RADIANS(__ANGLE__) (((__ANGLE__) / 180.0) * M_PI)
#define RADIANS_TO_DEGREES(__ANGLE__) (((__ANGLE__) / M_PI) * 180.0)

typedef struct
{
	GLfloat x;
	GLfloat y;
	GLfloat z;
} Vector3D;

typedef struct
{
	GLfloat x;
	GLfloat y;
} Vector2D;

typedef struct
{
	Vector3D	vertex;
	Vector3D	normal;
	Vector2D	texture;
} TexturedVertex3D;

static inline void substractVectors3D(Vector3D *a, Vector3D *b, Vector3D *result)
{
	result->x = a->x - b->x;
	result->y = a->y - b->y;
	result->z = a->z - b->z;
}

static inline void substractVectors2D(Vector2D *a, Vector2D *b, Vector2D *result)
{
	result->x = a->x - b->x;
	result->y = a->y - b->y;
}

static inline void negateVectors3D(Vector3D *a, Vector3D *minusa)
{
	minusa->x = -a->x;
	minusa->y = -a->y;
	minusa->z = -a->z;
}

static inline void negateVectors2D(Vector2D *a, Vector2D *minusa)
{
	minusa->x = -a->x;
	minusa->y = -a->y;
}

static inline GLfloat vectorLength3D(Vector3D *v)
{
	return sqrtf(v->x * v->x + v->y * v->y + v->z * v->z);
}

static inline GLfloat vectorLength2D(Vector2D *v)
{
	return sqrtf(v->x * v->x + v->y * v->y);
}

static inline void normalizeVector3D(Vector3D *v)
{
	GLfloat length = vectorLength3D(v);
	v->x /= length;
	v->y /= length;
	v->z /= length;
}

static inline void normalizeVector2D(Vector2D *v)
{
	GLfloat length = vectorLength2D(v);
	v->x /= length;
	v->y /= length;
}

static inline void multiplyVector3D(Vector3D *src, Vector3D *dst, CGFloat factor)
{
	dst->x = src->x * factor;
	dst->y = src->y * factor;
	dst->z = src->z * factor;
}

static inline void multiplyVector2D(Vector2D *src, Vector2D *dst, CGFloat factor)
{
	dst->x = src->x * factor;
	dst->y = src->y * factor;
}

static inline void divideVector3D(Vector3D *src, Vector3D *dst, CGFloat factor)
{
	multiplyVector3D(src, dst, 1/factor);
}

static inline void divideVector2D(Vector2D *src, Vector2D *dst, CGFloat factor)
{
	multiplyVector2D(src, dst, 1/factor);
}

static inline void crossProduct(Vector3D *a, Vector3D *b, Vector3D *result)
{
	result->x = a->y * b->z - a->z * b->y;
	result->y = a->z * b->x - a->x * b->z;
	result->z = a->x * b->y - a->y * b->x;
}	

static inline GLfloat dotProduct3D(Vector3D *a, Vector3D *b)
{
	return a->x * b->x + a->y * b->y + a->z * b->z;
}

static inline GLfloat dotProduct2D(Vector2D *a, Vector2D *b)
{
	return a->x * b->x + a->y * b->y;
}

static inline void vectorAngles(Vector3D *a, Vector3D *b, Vector3D *up, GLfloat *azimuth, GLfloat *elevation)
{
	CGFloat alen = vectorLength3D(a);
	CGFloat blen = vectorLength3D(b);
	CGFloat uplen = vectorLength3D(up);
	
	// Project vectors on UP
	CGFloat aup = dotProduct3D(a, up);
	CGFloat bup = dotProduct3D(b, up);
	GLfloat a_elevation = asin(aup/(alen * uplen));
	GLfloat b_elevation = asin(bup/(blen * uplen));
	*elevation = a_elevation - b_elevation;
	
	// Project vectors on plane orthogonal to UP
	// by substracting vectors probjection on UP from itself
	Vector3D aupvec, bupvec, aplane, bplane;
	// Vector projections on up:
	multiplyVector3D(up, &aupvec, aup / (uplen * uplen));
	multiplyVector3D(up, &bupvec, bup / (uplen * uplen));
	substractVectors3D(a, &aupvec, &aplane);
	substractVectors3D(b, &bupvec, &bplane);
	*azimuth = acos(dotProduct3D(&aplane, &bplane)/(vectorLength3D(&aplane)*vectorLength3D(&bplane)));
	// Now we only have the absolute value for azimuth, must get sign
	// For that, compute aplane x bplane, dot multiply it with up and see the sign
	Vector3D apcrossbp;
	crossProduct(&aplane, &bplane, &apcrossbp);
	if(dotProduct3D(&apcrossbp, up) > 0.0)
	{
		*azimuth = -*azimuth;
	}
	*elevation = RADIANS_TO_DEGREES(*elevation);
	*azimuth = RADIANS_TO_DEGREES(*azimuth);
}

static inline void matrixFillDefaultBottomRow(GLfloat matrix[][4])
{
	matrix[0][3] = 0.0;
	matrix[1][3] = 0.0;
	matrix[2][3] = 0.0;	
	matrix[3][3] = 1.0;
}
	
static inline void matrixFillDefault4thColumnAndRow(GLfloat matrix[][4])
{
	matrix[3][0] = 0.0;
	matrix[3][1] = 0.0;
	matrix[3][2] = 0.0;
	matrixFillDefaultBottomRow(matrix);
}

static inline void matrixAddTranslation(GLfloat matrix[][4], Vector3D *trans)
{
	matrix[3][0] = trans->x;
	matrix[3][1] = trans->y;
	matrix[3][2] = trans->z;
}

static inline void matrixCreateFromColumns(GLfloat result[][4], Vector3D *col0, Vector3D *col1, Vector3D *col2)
{
	result[0][0] = col0->x;
	result[0][1] = col0->y;
	result[0][2] = col0->z;
	result[1][0] = col1->x;
	result[1][1] = col1->y;
	result[1][2] = col1->z;
	result[2][0] = col2->x;
	result[2][1] = col2->y;
	result[2][2] = col2->z;
	matrixFillDefault4thColumnAndRow(result);
}

static inline void matrixCreateFromColumnsWithTranslation(GLfloat result[][4], Vector3D *col0, Vector3D *col1, Vector3D *col2, Vector3D *trans)
{
	result[0][0] = col0->x;
	result[0][1] = col0->y;
	result[0][2] = col0->z;
	result[1][0] = col1->x;
	result[1][1] = col1->y;
	result[1][2] = col1->z;
	result[2][0] = col2->x;
	result[2][1] = col2->y;
	result[2][2] = col2->z;
	matrixAddTranslation(result, trans);
	matrixFillDefaultBottomRow(result);
}

static inline void matrixCreateFromRows(GLfloat result[][4], Vector3D *col0, Vector3D *col1, Vector3D *col2)
{
	result[0][0] = col0->x;
	result[1][0] = col0->y;
	result[2][0] = col0->z;
	result[0][1] = col1->x;
	result[1][1] = col1->y;
	result[2][1] = col1->z;
	result[0][2] = col2->x;
	result[1][2] = col2->y;
	result[2][2] = col2->z;
	matrixFillDefault4thColumnAndRow(result);
}

static inline void matrixCreateFromRowsWithTranslation(GLfloat result[][4], Vector3D *col0, Vector3D *col1, Vector3D *col2, Vector3D *trans)
{
	result[0][0] = col0->x;
	result[1][0] = col0->y;
	result[2][0] = col0->z;
	result[0][1] = col1->x;
	result[1][1] = col1->y;
	result[2][1] = col1->z;
	result[0][2] = col2->x;
	result[1][2] = col2->y;
	result[2][2] = col2->z;
	matrixAddTranslation(result, trans);
	matrixFillDefault4thColumnAndRow(result);
}

#define vx(buffer) buffer[0]
#define vy(buffer) buffer[1]
#define vz(buffer) buffer[3]
static inline void setVector3D(Vector3D *v, GLfloat x, GLfloat y, GLfloat z)
{
	v->x = x;
	v->y = y;
	v->z = z;
}

static inline void setVector2D(Vector2D *v, GLfloat x, GLfloat y)
{
	v->x = x;
	v->y = y;
}

static inline void copyVector3D(Vector3D *src, Vector3D *dst)
{
	dst->x = src->x;
	dst->y = src->y;
	dst->z = src->z;
}

static inline void copyVector2D(Vector2D *src, Vector2D *dst)
{
	dst->x = src->x;
	dst->y = src->y;
}

// This is a modified version of the function of the same name from 
// the Mesa3D project ( http://mesa3d.org/ ), which is  licensed
// under the MIT license, which allows use, modification, and 
// redistribution
static inline void gluLookAt(GLfloat eyex, GLfloat eyey, GLfloat eyez,
							 GLfloat centerx, GLfloat centery, GLfloat centerz,
							 GLfloat upx, GLfloat upy, GLfloat upz)
{
	GLfloat m[16];
	GLfloat x[3], y[3], z[3];
	GLfloat mag;
	
	/* Make rotation matrix */
	
	/* Z vector */
	z[0] = eyex - centerx;
	z[1] = eyey - centery;
	z[2] = eyez - centerz;
	mag = sqrtf(z[0] * z[0] + z[1] * z[1] + z[2] * z[2]);
	if (mag) {			/* mpichler, 19950515 */
		z[0] /= mag;
		z[1] /= mag;
		z[2] /= mag;
	}
	
	/* Y vector */
	y[0] = upx;
	y[1] = upy;
	y[2] = upz;
	
	/* X vector = Y cross Z */
	x[0] = y[1] * z[2] - y[2] * z[1];
	x[1] = -y[0] * z[2] + y[2] * z[0];
	x[2] = y[0] * z[1] - y[1] * z[0];
	
	/* Recompute Y = Z cross X */
	y[0] = z[1] * x[2] - z[2] * x[1];
	y[1] = -z[0] * x[2] + z[2] * x[0];
	y[2] = z[0] * x[1] - z[1] * x[0];
	
	/* mpichler, 19950515 */
	/* cross product gives area of parallelogram, which is < 1.0 for
	 * non-perpendicular unit-length vectors; so normalize x, y here
	 */
	
	mag = sqrtf(x[0] * x[0] + x[1] * x[1] + x[2] * x[2]);
	if (mag) {
		x[0] /= mag;
		x[1] /= mag;
		x[2] /= mag;
	}
	
	mag = sqrtf(y[0] * y[0] + y[1] * y[1] + y[2] * y[2]);
	if (mag) {
		y[0] /= mag;
		y[1] /= mag;
		y[2] /= mag;
	}
	
#define M(row,col)  m[col*4+row]
	M(0, 0) = x[0];
	M(0, 1) = x[1];
	M(0, 2) = x[2];
	M(0, 3) = 0.0;
	M(1, 0) = y[0];
	M(1, 1) = y[1];
	M(1, 2) = y[2];
	M(1, 3) = 0.0;
	M(2, 0) = z[0];
	M(2, 1) = z[1];
	M(2, 2) = z[2];
	M(2, 3) = 0.0;
	M(3, 0) = 0.0;
	M(3, 1) = 0.0;
	M(3, 2) = 0.0;
	M(3, 3) = 1.0;
#undef M
	glMultMatrixf(m);
	
	/* Translate Eye to Origin */
	glTranslatef(-eyex, -eyey, -eyez);
	
}