//
//  VARObject.m
//  VirtuAR
//
//  Created by Romain on 04/02/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import "VARGLutils.h"
#import "VARObject.h"
#import "VARCamera.h"

@implementation VARObject

@synthesize cam;
@synthesize distanceToCam;
@synthesize cartesianPos;
@synthesize vectorUp;
@synthesize vectorFront;

-(id)init
{
	if(self = [super init])
	{
		mustUpdateDistanceToCam = NO;
		previousPositionChecksum = 0.0;
	}
	
	return self;
}

-(void)draw:(NSTimeInterval) timestamp
{
	
}

-(void)setCartesianPosWithX:(GLfloat)x Y:(GLfloat)y Z:(GLfloat)z
{
	cartesianPos.x = x;
	cartesianPos.y = y;
	cartesianPos.z = z;
	
	mustUpdateDistanceToCam = YES;
}

-(void)setCartesianPos2DWithX:(GLfloat)x Y:(GLfloat)y
{
	cartesianPos.x = x;
	cartesianPos.y = y;

	mustUpdateDistanceToCam = YES;
}

-(void)setCartesianPos:(Vector3D)pos
{
	cartesianPos.x = pos.x;
	cartesianPos.y = pos.y;
	cartesianPos.z = pos.z;

	mustUpdateDistanceToCam = YES;
}

-(void)setCartesianPos2D:(Vector2D)pos
{
	cartesianPos.x = pos.x;
	cartesianPos.y = pos.y;

	mustUpdateDistanceToCam = YES;
}

-(void)setCartesianPosFromPointer:(Vector3D *)pos
{
	cartesianPos.x = pos->x;
	cartesianPos.y = pos->y;
	cartesianPos.z = pos->z;

	mustUpdateDistanceToCam = YES;
}

-(void)setCartesianPos2DFromPointer:(Vector2D *)pos
{
	cartesianPos.x = pos->x;
	cartesianPos.y = pos->y;

	mustUpdateDistanceToCam = YES;
}

-(void)setAltitude:(CLLocationDistance)alt
{
	cartesianPos.z = alt;

	mustUpdateDistanceToCam = YES;
}

-(CLLocationDistance)altitude
{
	return cartesianPos.z;
}

-(void)setVectorUpWithX:(GLfloat)x Y:(GLfloat)y Z:(GLfloat)z
{
	vectorUp.x = x;
	vectorUp.y = y;
	vectorUp.z = z;

	mustUpdateDistanceToCam = YES;
}

-(void)setVectorUpFromPointer:(Vector3D *)up
{
	vectorUp.x = up->x;
	vectorUp.y = up->y;
	vectorUp.z = up->z;

	mustUpdateDistanceToCam = YES;
}

-(void)setVectorFrontWithX:(GLfloat)x Y:(GLfloat)y Z:(GLfloat)z
{
	vectorFront.x = x;
	vectorFront.y = y;
	vectorFront.z = z;

	mustUpdateDistanceToCam = YES;
}

-(void)setVectorFrontFromPointer:(Vector3D *)front;
{
	vectorFront.x = front->x;
	vectorFront.y = front->y;
	vectorFront.z = front->z;

	mustUpdateDistanceToCam = YES;
}

-(CLLocationDistance)distanceToCam
{
	if(mustUpdateDistanceToCam || !ARE_FLOAT_EQUALS(previousPositionChecksum, cam.positionChecksum))
	{
		[self _updateDistanceToCam];
		mustUpdateDistanceToCam = NO;
		previousPositionChecksum = cam.positionChecksum;
	}
	
	return distanceToCam;
}

-(void)_updateDistanceToCam
{
	distanceToCam = [self distanceToObject:self.cam];
}

-(void)setCam:(VARCamera *)camera
{
	[camera retain];
	[cam release];
	cam = camera;
	
	mustUpdateDistanceToCam = YES;	
}

-(void)updateOrientationMatrix
{
	normalizeVector3D(&vectorUp);
	normalizeVector3D(&vectorFront);
	
	// Create transformation matrix
	Vector3D thirdVector; // = Up x Front
	crossProduct(&vectorUp, &vectorFront, &thirdVector);
	
	matrixCreateFromRows(orientationMatrix, &vectorFront, &thirdVector, &vectorUp);
}
-(void)updateOrientationMatrixWithUp:(Vector3D)up Front:(Vector3D)front
{
	self.vectorUp = up;
	self.vectorFront = front;
	[self updateOrientationMatrix];
}
	
-(void)updateOrientationMatrixWithPointersUp:(Vector3D *)up Front:(Vector3D *)front
{
	[self setVectorUpFromPointer:up];
	[self setVectorFrontFromPointer:front];
	[self updateOrientationMatrix];
}

-(GLfloat)distanceToObject:(VARObject *)obj
{
	Vector3D delta;
	Vector3D objPos = obj.cartesianPos;
	substractVectors3D(&cartesianPos, &objPos, &delta);
	return vectorLength3D(&delta);
}

#pragma mark MWLinkedListCompare methods
-(int)compareToObject:(id)obj
{
	if([obj respondsToSelector:@selector(distanceToCam)])
	{
		VARObject *tmp = (VARObject *)obj;
		return (int)(self.distanceToCam - tmp.distanceToCam);
	}
	else
	{
		return 0;
	}
}

-(void)dealloc
{
	[cam release];
	[super dealloc];
}

@end
