//
//  VARSphere.m
//  videoGL
//
//  Created by Romain on 29/04/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import "VARSphere.h"

#define DEFAULT_PRECISION	16
#define PI 3.141592654
#define TWOPI 6.283185308
#define PID2	(PI/2)

int MakeSphere(TexturedVertex3D **buffer, Vector3D c, GLfloat r, int n)
{
	int i,j;
	double theta1,theta2,theta3;
	Vector3D e,p;
	
	if (r < 0)
		r = -r;
	if (n < 0)
		n = -n;

	if (n < 4 || r <= 0)
	{
		return 0;
	}
	
	int numVertices = 2 * ((n/2) * (n + 1));
	*buffer = calloc(numVertices, sizeof(TexturedVertex3D));
	TexturedVertex3D *currentVertex = *buffer;
	
	for (j=0;j<n/2;j++) {
		theta1 = j * TWOPI / n - PID2;
		theta2 = (j + 1) * TWOPI / n - PID2;
		
		for (i=0;i<=n;i++) {
			theta3 = i * TWOPI / n;
			
			e.x = cos(theta2) * cos(theta3);
			e.y = sin(theta2);
			e.z = cos(theta2) * sin(theta3);
			p.x = c.x + r * e.x;
			p.y = c.y + r * e.y;
			p.z = c.z + r * e.z;
			
			
			currentVertex->vertex = p;
			setVector3D(&(currentVertex->normal), -e.x, -e.y, -e.z);
			setVector2D(&(currentVertex->texture), i/(double)n,2*(j+1)/(double)n);
			currentVertex++;
			
			e.x = cos(theta1) * cos(theta3);
			e.y = sin(theta1);
			e.z = cos(theta1) * sin(theta3);
			p.x = c.x + r * e.x;
			p.y = c.y + r * e.y;
			p.z = c.z + r * e.z;
			
			currentVertex->vertex = p;
			setVector3D(&(currentVertex->normal), -e.x, -e.y, -e.z);
			setVector2D(&(currentVertex->texture), i/(double)n,2*j/(double)n);
			currentVertex++;
		}
	}
	return numVertices;
}

@implementation VARSphere
@synthesize radius = _radius;
@synthesize texture = _texture;
@synthesize precision = _precision;

-(id)initWithRadius:(GLfloat)r
{
	return [self initWithRadius:r precision:DEFAULT_PRECISION];
}

-(id)initWithRadius:(GLfloat)r precision:(int)n
{
	if(self = [super init])
	{
		Vector3D center;
		center.x = center.y = center.z = 0.0;
		_vertexData = NULL;
		_numVertices = MakeSphere(&_vertexData, center, r, n);
		_radius = r;
		_precision = n;
		setVector3D(&vectorUp, 0.0, -1.0, 0.0);
		setVector3D(&vectorFront, 0.0, 0.0, 1.0);
		[self updateOrientationMatrix];
	}
	return self;
}

-(void)draw:(NSTimeInterval) timestamp
{
	if(_texture == nil)
		return;
	
	glPushMatrix();
	
	// Translate object to its position
	glTranslatef(cartesianPos.x, cartesianPos.y, cartesianPos.z);
	
	// Set correct object orientation:
	// - Up is Z
	// - Front is X
	glMultMatrixf((GLfloat*)orientationMatrix);
	
	glBlendFunc(GL_ONE, GL_SRC_COLOR);	
	//glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	[_texture bind];
	
	glVertexPointer(3, GL_FLOAT, sizeof(TexturedVertex3D), &_vertexData[0].vertex);
    glNormalPointer(GL_FLOAT, sizeof(TexturedVertex3D), &_vertexData[0].normal);
    glTexCoordPointer(2, GL_FLOAT, sizeof(TexturedVertex3D), &_vertexData[0].texture);
	
    glDrawArrays(GL_TRIANGLE_STRIP, 0, _numVertices);
	
	glPopMatrix();
}

/*
 Create a sphere centered at c, with radius r, and precision n
 Draw a point for zero radius spheres
 */
/*
void CreateSphere(XYZ c,double r,int n)
{
	int i,j;
	double theta1,theta2,theta3;
	XYZ e,p;
	
	if (r < 0)
		r = -r;
	if (n < 0)
		n = -n;
	if (n < 4 || r <= 0) {
		glBegin(GL_POINTS);
		glVertex3f(c.x,c.y,c.z);
		glEnd();
		return;
	}
	
	for (j=0;j<n/2;j++) {
		theta1 = j * TWOPI / n - PID2;
		theta2 = (j + 1) * TWOPI / n - PID2;
		
		glBegin(GL_QUAD_STRIP);
		for (i=0;i<=n;i++) {
			theta3 = i * TWOPI / n;
			
			e.x = cos(theta2) * cos(theta3);
			e.y = sin(theta2);
			e.z = cos(theta2) * sin(theta3);
			p.x = c.x + r * e.x;
			p.y = c.y + r * e.y;
			p.z = c.z + r * e.z;
			
			glNormal3f(e.x,e.y,e.z);
			glTexCoord2f(i/(double)n,2*(j+1)/(double)n);
			glVertex3f(p.x,p.y,p.z);
			
			e.x = cos(theta1) * cos(theta3);
			e.y = sin(theta1);
			e.z = cos(theta1) * sin(theta3);
			p.x = c.x + r * e.x;
			p.y = c.y + r * e.y;
			p.z = c.z + r * e.z;
			
			glNormal3f(e.x,e.y,e.z);
			glTexCoord2f(i/(double)n,2*j/(double)n);
			glVertex3f(p.x,p.y,p.z);
		}
		glEnd();
	}
}
*/
/*
 Create a sphere centered at c, with radius r, and precision n
 Draw a point for zero radius spheres
 Use CCW facet ordering
 "method" is 0 for quads, 1 for triangles
 (quads look nicer in wireframe mode)
 Partial spheres can be created using theta1->theta2, phi1->phi2
 in radians 0 < theta < 2pi, -pi/2 < phi < pi/2
 */
/*
void CreateSphere(XYZ c,double r,int n,int method,
				  double theta1,double theta2,double phi1,double phi2)
{
	int i,j;
	double t1,t2,t3;
	XYZ e,p;
	
	// Handle special cases
	if (r < 0)
		r = -r;
	if (n < 0)
		n = -n;
	if (n < 4 || r <= 0) {
		glBegin(GL_POINTS);
		glVertex3f(c.x,c.y,c.z);
		glEnd();
		return;
	}
	
	for (j=0;j<n/2;j++) {
		t1 = phi1 + j * (phi2 - phi1) / (n/2);
		t2 = phi1 + (j + 1) * (phi2 - phi1) / (n/2);
		
		if (method == 0)
			glBegin(GL_QUAD_STRIP);
		else
			glBegin(GL_TRIANGLE_STRIP);
		
		for (i=0;i<=n;i++) {
			t3 = theta1 + i * (theta2 - theta1) / n;
			
			e.x = cos(t1) * cos(t3);
			e.y = sin(t1);
			e.z = cos(t1) * sin(t3);
			p.x = c.x + r * e.x;
			p.y = c.y + r * e.y;
			p.z = c.z + r * e.z;
			glNormal3f(e.x,e.y,e.z);
			glTexCoord2f(i/(double)n,2*j/(double)n);
			glVertex3f(p.x,p.y,p.z);
			
			e.x = cos(t2) * cos(t3);
			e.y = sin(t2);
			e.z = cos(t2) * sin(t3);
			p.x = c.x + r * e.x;
			p.y = c.y + r * e.y;
			p.z = c.z + r * e.z;
			glNormal3f(e.x,e.y,e.z);
			glTexCoord2f(i/(double)n,2*(j+1)/(double)n);
			glVertex3f(p.x,p.y,p.z);
			
		}
		glEnd();
	}
}
*/

@end
