//
//  VARTexture.m
//  VirtuAR
//
//  Created by Romain on 01/03/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import "VARTexture.h"

static GLuint _currentTexture = 0;

@interface VARTexture()
-(void)_bindIfNeeded;	
@end

@implementation VARTexture
@synthesize width;
@synthesize height;
@synthesize wrapS;
@synthesize wrapT;
@synthesize minFilter;
@synthesize magFilter;

-(id)init
{
	if(self = [super init])
	{
		wrapS = GL_CLAMP_TO_EDGE;
		wrapT = GL_CLAMP_TO_EDGE;
		minFilter = GL_LINEAR;
		magFilter = GL_LINEAR;
		glGenTextures(1, &name[0]);
	}
	
	return self;
}

-(void)setDataFromUIImage:(UIImage *)img
{
	[self updateDataFromUIImage:img];
}

-(void)updateDataFromUIImage:(UIImage *)img
{
	if (img == nil)
	{
        NSLog(@"Do real error checking here");
		return;
	}
	
	[self _bindIfNeeded];
	
    GLuint w = CGImageGetWidth(img.CGImage);
    GLuint h = CGImageGetHeight(img.CGImage);
    CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
    void *textureData = malloc( h * w * 4 );
    CGContextRef context = CGBitmapContextCreate( textureData, w, h, 8, 4 * w, colorSpace, kCGImageAlphaPremultipliedLast | kCGBitmapByteOrder32Big );
	CGContextTranslateCTM (context, 0, h);
	CGContextScaleCTM (context, 1.0, -1.0);
    CGColorSpaceRelease( colorSpace );
	
	//    CGContextTranslateCTM( context, 0, h);
	
    CGContextClearRect( context, CGRectMake( 0, 0, w, h ) );
    CGContextDrawImage( context, CGRectMake( 0, 0, w, h ), img.CGImage );
    CGContextRelease(context);
	
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, textureData);
	free(textureData);
	width = w;
	height = h;
}

-(void)setDataFromRaw:(void *)img width:(GLuint)w height:(GLuint)h
{
	[self updateDataFromRaw:img width:w height:h];
}

-(void)updateDataFromRaw:(void *)img width:(GLuint)w height:(GLuint)h
{
	if (img == nil)
	{
        NSLog(@"Do real error checking here");
		return;
	}
	
	[self _bindIfNeeded];
	
#if TARGET_IPHONE_SIMULATOR
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, w, h, 0, GL_RGB, GL_UNSIGNED_BYTE, img);
#else
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, img);
#endif
//	free(img);
	width = w;
	height = h;
}

-(void)bind
{
	[self _bindIfNeeded];
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,self.minFilter); 
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,self.magFilter);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, self.wrapS);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, self.wrapT);
}

-(void)_bindIfNeeded
{
	if(name[0] != _currentTexture)
	{
		glBindTexture(GL_TEXTURE_2D, name[0]);
		_currentTexture = name[0];
	}
}

-(void)redraw
{
}

-(void)dealloc
{
	glDeleteTextures(1, &name[0]);
	[super dealloc];
}

@end
