//
//  VARViewTexture.m
//  VirtuAR
//
//  Created by Romain on 01/03/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//
#import <UIKit/UIKit.h>
#import "VARViewTexture.h"
#import <QuartzCore/QuartzCore.h>

typedef struct
{
	void *buffer;
	GLuint width;
	GLuint height;
} rawImageBuffer;

static NSThread *_textureThread;

@interface VARViewTexture()
+(void)_textureThreadMain:(id)arg;
-(void)_drawTexture;
-(void)_setDataFromRaw:(NSData *)buf;
@end


@implementation VARViewTexture
@synthesize view;

+(void)initialize
{
	if(self == [VARViewTexture class])
	{
		_textureThread = [[NSThread alloc] initWithTarget:self
											 selector:@selector(_textureThreadMain:)
											   object:nil];
		[_textureThread start];  // Actually create the thread
	}
}

-(id)initWithView:(UIView *)v
{
	if(self = [super init])
	{
		[self setView:v];
	}
	return self;
}

-(void)setView:(UIView *)v
{
	[v retain];
	[view release];
	view = v;
	
	[self redraw];
}

-(void)redraw
{
	_callerThread = [NSThread currentThread];
	[self performSelector:@selector(_drawTexture) onThread:_textureThread withObject:nil waitUntilDone:NO];
}

-(void)_drawTexture
{
	CGRect rect = view.bounds;
	CGContextRef context;
// Original View to texture rendering code
// Rendered the view to a UIImage, which had to be rendered then to a raw buffer
/*
	UIGraphicsBeginImageContext(rect.size);
	ctx = UIGraphicsGetCurrentContext();
	[view.layer renderInContext:ctx];
	UIImage *viewImage = UIGraphicsGetImageFromCurrentImageContext();
	UIGraphicsEndImageContext();
	
	[self setDataFromUIImage:viewImage];
*/
	GLuint w = (GLuint)rect.size.width;
    GLuint h = (GLuint)rect.size.height;
    CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
    void *textureData = malloc( h * w * 4 );
	context = CGBitmapContextCreate( textureData, w, h, 8, 4 * w, colorSpace, kCGImageAlphaPremultipliedLast | kCGBitmapByteOrder32Big );
//	CGContextTranslateCTM (context, 0, h);
//	CGContextScaleCTM (context, 1.0, -1.0);
    CGColorSpaceRelease( colorSpace );
	
    CGContextClearRect( context, CGRectMake( 0, 0, w, h ) );
	[view.layer renderInContext:context];
    CGContextRelease(context);

	rawImageBuffer *buf = malloc(sizeof(rawImageBuffer));
	buf->buffer = textureData;
	buf->width = w;
	buf->height = h;
	[self performSelector:@selector(_setDataFromRaw:) onThread:_callerThread withObject:[NSData dataWithBytesNoCopy:buf length:sizeof(rawImageBuffer)] waitUntilDone:NO];
}

-(void)_setDataFromRaw:(NSData *)buf
{
	rawImageBuffer *img = (rawImageBuffer *)[buf bytes];
	[self setDataFromRaw:img->buffer width:img->width height:img->height];
}

-(void)bind
{
	[super bind];
}


+(void)_textureThreadMain:(id)arg
{
	[NSThread setThreadPriority:0.4];
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	
	//A runloop with no sources returns immediately from runMode:beforeDate:
	//That will wake up the loop and chew CPU. Add a dummy source to prevent
	//it.
	
	NSRunLoop *runLoop = [NSRunLoop currentRunLoop];
	
	NSMachPort *dummyPort = [[NSMachPort alloc] init];
	[runLoop addPort:dummyPort forMode:NSDefaultRunLoopMode];
	[dummyPort release];
	[pool release];
	
	while (1) 
	{
		NSAutoreleasePool *loopPool = [[NSAutoreleasePool alloc] init];
		[runLoop runMode:NSDefaultRunLoopMode beforeDate:[NSDate distantFuture]];
//		NSLog(@"VARViewTexture: finished Run Loop");
		[loopPool drain];
	}
}
-(void)dealloc
{
	[view release];
	[super dealloc];
}

@end
