//
//  VARWorld.m
//  VirtuAR
//
//  Created by Romain on 04/02/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import "VARWorld.h"

@implementation VARWorld
@synthesize camera;
@synthesize objectList;
@synthesize drawDistance = _drawDistance;
@synthesize maxObjectCount = _maxObjectCount;

// Create an ES 1.1 context
- (id) init
{
	if (self = [super init])
	{
		context = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES1];
        
        if (!context || ![EAGLContext setCurrentContext:context])
		{
            [self release];
            return nil;
        }
		
		// Create default framebuffer object. The backing will be allocated for the current layer in -resizeFromLayer
		glGenFramebuffersOES(1, &defaultFramebuffer);
		glGenRenderbuffersOES(1, &colorRenderbuffer);
		glBindFramebufferOES(GL_FRAMEBUFFER_OES, defaultFramebuffer);
		glBindRenderbufferOES(GL_RENDERBUFFER_OES, colorRenderbuffer);
		glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES, GL_COLOR_ATTACHMENT0_OES, GL_RENDERBUFFER_OES, colorRenderbuffer);

		/*
		glGenRenderbuffersOES(1, &depthRenderbuffer);
		glBindRenderbufferOES(GL_RENDERBUFFER_OES, depthRenderbuffer);
		glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES, GL_DEPTH_ATTACHMENT_OES, GL_RENDERBUFFER_OES, depthRenderbuffer);
		*/
		 
		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glEnableClientState(GL_NORMAL_ARRAY);
		glEnable(GL_TEXTURE_2D);
//		glEnable(GL_BLEND);
//		glEnable(GL_DEPTH_TEST);
//		glEnable(GL_NORMALIZE);
		glDepthFunc( GL_LEQUAL     );
		glDepthMask( GL_TRUE );
		
//		glShadeModel(GL_FLAT);
		
		glEnable(GL_LIGHTING);
		
		// Turn the first light on
		glEnable(GL_LIGHT0);
		
		// Define the ambient component of the first light
		static const GLfloat light0Ambient[] = {0.4, 0.4, 0.4, 1.0};
		glLightfv(GL_LIGHT0, GL_AMBIENT, light0Ambient);
		
		// Define the diffuse component of the first light
		static const GLfloat light0Diffuse[] = {0.8, 0.8, 0.8, 1.0};
		glLightfv(GL_LIGHT0, GL_DIFFUSE, light0Diffuse);
		
		// Define the position of the first light
		// const GLfloat light0Position[] = {10.0, 10.0, 10.0}; 
		// static const GLfloat light0Position[] = {10.0, 10.0, 10.0};
		// glLightfv(GL_LIGHT0, GL_POSITION, light0Position); 

		camera = [[VARCamera alloc] init];
		camera.altitude = 1.7;
		_drawDistance = camera.zFar = 2000.0;
		_maxObjectCount = 256;
		
		self.objectList = [[MWLinkedList alloc] init];
		objectList.sortType = MWLL_SORTED_DESCENDING;
/*
		horizon = [[VAR3DSprite alloc] init];
		[horizon setTextureFileName:@"gridtile"];
		horizon.textureWrapS = GL_REPEAT;
		horizon.textureWrapT = GL_REPEAT;
		[horizon setSize:100.0];
		[horizon setTextureScale:25.0];
		[horizon setBillboard:NO];
		
		compass = [[VAR3DSprite alloc] init];
		[compass setTextureFileName:@"compasstexture"];
		[compass setSize:10.0];
		[compass setTextureScale:1.0];
		[compass setBillboard:NO];
		[compass setAltitude:0.01];
*/
	}
	
	return self;
}

-(void)setDrawDistance:(GLfloat)dist
{
	_drawDistance = dist;
	camera.zFar = dist * 1.01;
}

-(void)addObject:(VARObject *)obj
{
	[objectList insertSortedObject:obj];
}

-(void)removeObject:(VARObject *)obj
{
	[objectList removeObject:obj];
}

-(void)setLocation:(Vector2D *)pos altitude:(CLLocationDistance)alt direction:(CLLocationDirection)dir acceleration:(Vector3D *)acc;
{
	if(pos != NULL)
	{
		copyVector2D(pos, &position);
		positionChanged = YES;
	}

	direction = dir;
	copyVector3D(acc, &accel);
}

-(void)render:(NSTimeInterval)tstamp
{
	// This application only creates a single context which is already set current at this point.
	// This call is redundant, but needed if dealing with multiple contexts.
    [EAGLContext setCurrentContext:context];
    
	// This application only creates a single default framebuffer which is already bound at this point.
	// This call is redundant, but needed if dealing with multiple framebuffers.
    glBindFramebufferOES(GL_FRAMEBUFFER_OES, defaultFramebuffer);
	
    glClearColor(0.5f, 0.5f, 0.5f, 0.0f);
//    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
//    glClear(GL_COLOR_BUFFER_BIT);
	
	glPushMatrix();
	
	glPushMatrix();
//	glTranslatef(cartesianPosition.x, cartesianPosition.y, 10.0);
	static const GLfloat light0Position[] = {10.0, 10.0, 10.0, 10.0};
	glLightfv(GL_LIGHT0, GL_POSITION, light0Position);
	glPopMatrix();
	
	if(positionChanged)
	{
		[horizon setCartesianPos2DFromPointer:&position];
		[compass setCartesianPos2DFromPointer:&position];
		[objectList sort];
		positionChanged = NO;
	}
	
	[camera setAccel:accel];
	[camera setDirection:direction];
	
	[camera draw:tstamp];
//	[horizon draw:tstamp];
//	[compass draw:tstamp];
	
	VARObject *curObject = nil;

	[objectList rewindCursor];
	uint remainingObjectCount = [objectList count];
	// Walk the object list until we find the 1st that's within visibility range
	// or until there are less than maxObjectCount remaining
	while((curObject = (VARObject *)[objectList nextObject]) != nil)
	{	
		if((curObject.distanceToCam < _drawDistance) && (remainingObjectCount <= _maxObjectCount))
		{
			[curObject draw:tstamp];
			break;
		}
		remainingObjectCount--;
	}
	// Continue walking the list, no need to test the draw distance
	// because the list is sorted in descending order
	while((curObject = (VARObject *)[objectList nextObject]) != nil)
	{	
		[curObject draw:tstamp];
	}
	glPopMatrix();
	
	// This application only creates a single color renderbuffer which is already bound at this point.
	// This call is redundant, but needed if dealing with multiple renderbuffers.
    glBindRenderbufferOES(GL_RENDERBUFFER_OES, colorRenderbuffer);
    [context presentRenderbuffer:GL_RENDERBUFFER_OES];
}

- (BOOL) resizeFromLayer:(CAEAGLLayer *)layer
{	
	// Allocate color buffer backing based on the current layer size
    glBindRenderbufferOES(GL_RENDERBUFFER_OES, colorRenderbuffer);
    [context renderbufferStorage:GL_RENDERBUFFER_OES fromDrawable:layer];
	glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_WIDTH_OES, &backingWidth);
    glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_HEIGHT_OES, &backingHeight);
	
/*
	glBindRenderbufferOES(GL_RENDERBUFFER_OES, depthRenderbuffer);
	glRenderbufferStorageOES(GL_RENDERBUFFER_OES, GL_DEPTH_COMPONENT16_OES, backingWidth, backingHeight);
*/
    if (glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES) != GL_FRAMEBUFFER_COMPLETE_OES)
	{
		NSLog(@"Failed to make complete framebuffer object %x", glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES));
        return NO;
    }
	
	[camera setViewSizeWithX:backingWidth Y:backingHeight];
    
    return YES;
}

- (void) dealloc
{
	// Tear down GL
	if (defaultFramebuffer)
	{
		glDeleteFramebuffersOES(1, &defaultFramebuffer);
		defaultFramebuffer = 0;
	}
	
	if (colorRenderbuffer)
	{
		glDeleteRenderbuffersOES(1, &colorRenderbuffer);
		colorRenderbuffer = 0;
	}

	if (depthRenderbuffer)
	{
		glDeleteRenderbuffersOES(1, &depthRenderbuffer);
		depthRenderbuffer = 0;
	}
	
	// Tear down context
	if ([EAGLContext currentContext] == context)
        [EAGLContext setCurrentContext:nil];
	
	[context release];
	context = nil;
	self.objectList = nil;
	[camera dealloc];
	
	[super dealloc];
}

@end
